"""
File IO for ASCII datafiles in fixed witdh format.
"""

from nilutility.datatypes import DataObject
import datetime
import logging

class DatetimeParserFactory(object):  # pylint: disable=R0903
    # R0903:Too few public methods
    """
    Factory class for datetime.datetime, constructed by parsing a string.
    """
    def __init__(self, formatstring):
        """
        Set up datetime_parser_factory.
        Parameters:
            None
        Returns:
            None
        """
        self.formatstring = formatstring

    def __call__(self, datetimestr):
        """
        Actually create the object: a datatime.datetime that is parsed from a
        string.
        Parameters:
            datetimestring     string containing the datetime value according to
                               the format string self.formatstr
        Returns:
            datetime.datetime object
        """
        return datetime.datetime.strptime(datetimestr, self.formatstring)


class FormatError(Exception):
    """
    File format error, raised during read.
    """
    pass


class NDL4(object):
    # pylint: disable=R0902
    # R0902: Too many instance attributes
    """
    File IO for NDL4 datafile format.

    Synopsis:

    General: setup object:
        ndl4 = NDL4()
        ndl4.add_column("varname1", int)
        ndl4.add_column("varname2")  # default is float
        try:
            ndl4.read("filename")
        except IOError as excpt:
            # handle this
            pass
        except FormatError as excpt:
            # handle that
            pass

    2 ways of usage/data access:

    1) access variables as lists:
        ndl4.start[0]  # first start time in file (as datetime.ddatetime)
        ndl4.start[-1]  # last end time in file (as datetime.ddatetime)
        ndl4.coverage  # data coverage list (last column)

        ndl4.varname1  # variable access as attibute (list)
        ndl4.data[0]   # same variable (list)

    2) access by starttime/end time
        # get the value for variable "varname1" for the sample time
        # 2018-01-01 12:36:00 - 2018-01-01 12:37:00
        ndl4.get("varname1",
                 datetime.datetime(2018, 1, 1, 12, 36, 0),
                 datetime.datetime(2018, 1, 1, 12, 37, 0))
    """

    def __init__(self):
        """
        Set up IO object.

        Parameters:
            None
        Returns:
            None
        Raises:
            None
        """
        self.logger = logging.getLogger('NDL4')
        self.format_desc = DataObject(columns=[])
        self.start = []
        self.end = []
        self.data = []
        self.coverage = []
        self.filename = None
        # faster access for sequential reads: remember last time and index
        self.recent_start = None  # last acccessed start time
        self.recent_index = None  # last acccessed index

    def add_column(self, name, datatype=float):
        """
        Adds a column to the format description.

        Parameters:
            pos_start   start position in file
            length      length of the column
            datatype    datatype to be used for the column
        """
        self.format_desc.columns.append(
            DataObject(name=name, datatype=datatype))
        self.data.append([])

    def __getattr__(self, name):
        """
        Attribute like access for variables: wrapper for var_by_name.
        Parameters:
            name   attribute name
        Returns:
            variable by name
        """
        try:
            return self.var_by_name(name)
        except ValueError:
            raise AttributeError("no column named '{}'".format(name))

    def var_by_name(self, name):
        """
        Get a variable by it's name.
        Parameters:
            name   variabble name
        Returns:
            variable by name
        """
        ind = [x.name for x in self.format_desc.columns].index(name)
        return self.data[ind]

    def read(self, filename):
        """
        Reads a file according to the data format description.

        Parameters:
            filename     file path and name

        Returns:
            None
        Raises:
            FormatError
        """
        self.filename = filename
        fil = open(self.filename)
        linenum = -1
        timeparser = DatetimeParserFactory('%Y.%m.%d %H:%M:%S')
        for line in fil:
            linenum += 1
            if len(line) < 39:
                raise FormatError('line {}: syntax error, too short line',
                                  linenum)
            # prosess start/end time
            try:
                self.start.append(timeparser(line[0:19]))
            except ValueError as excpt:
                raise FormatError('line {}, start time: {}', linenum,
                                  excpt.message)
            try:
                self.end.append(timeparser(line[20:39]))
            except ValueError as excpt:
                raise FormatError('line {}, end time: {}', linenum,
                                  excpt.message)
            # prosess data
            data = line[39:].split()
            if len(self.format_desc.columns)+3 != len(data)+2:
                raise FormatError(
                    'line {}: Data format specifies {} columns, {} found'
                    .format(linenum, len(self.format_desc.columns)+3,
                            len(data)+2))
            for var_index in range(len(self.format_desc.columns)):
                val = data[var_index]
                if val == "-9900":
                    val = None
                else:
                    val = self.format_desc.columns[var_index].datatype(val)
                self.data[var_index].append(val)
            var_index = len(self.format_desc.columns)  # last column is coverage
            self.coverage.append(int(data[var_index]))

    def get(self, name, start, end):
        """
        Get a data value by variable name and sample start/end times.
        Prameters:
            name    variable name
            start   sample start time of requested sample
            end     sample end time of requested sample
        Returns:
            One value for one variable at one sample interval.
            When the satrt/end time does not exists in the file, None is
            returned
        """
        ind = None
        if self.recent_start and self.recent_start == start:
            ind = self.recent_index
        else:
            startind = 0
            stopind = len(self.start)
            if self.recent_start and self.recent_start < start:
                startind = self.recent_index
            for i in range(startind, stopind):
                if self.start[i] == start:
                    ind = i
                    break
        if ind is not None and self.start[ind] == start and \
           self.end[ind] == end:
            if ind != self.recent_index:
                self.recent_start = start
                self.recent_index = ind
            return self.var_by_name(name)[ind]
        elif ind is None:
            errtxt = "Input file '{}': start {}, end {} not found".format(
                self.filename, start, end)
            self.logger.warning(errtxt)
            return None
        else:
            errtxt = ("Input file '{}', line {}: start {}, end {} does not "
                      "match requested start {}, end {}").format(
                          self.filename, ind+1, self.start[ind], self.end[ind],
                          start, end)
            self.logger.warning(errtxt)
            return None



from .generated.gml_generated import *
from .generated.gml_generated import quote_attrib, ExternalEncoding, showIndent, __all__

# correct version of export
# This could also be done by changing GeneratedsSuper, but this way all changes
# are coded in this file.
def _export(self, outfile, level, namespace_='gml:', name_='DirectPositionType', namespacedef_='xmlns:gml="http://www.opengis.net/gml/3.2"', pretty_print=True):
    if pretty_print:
        eol_ = '\n'
    else:
        eol_ = ''
    if self.original_tagname_ is not None:
        name_ = self.original_tagname_
    showIndent(outfile, level, pretty_print)
    outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
    already_processed = set()
    self.exportAttributes(outfile, level, already_processed, namespace_, name_='DirectPositionType')
    if self.hasContent_():
        outfile.write('>')
        res = []
        for val in self.valueOf_:
            res.append(str(val).encode(ExternalEncoding))
        outfile.write(' '.join(res))
        self.exportChildren(outfile, level + 1, namespace_='gml:', name_='DirectPositionType', pretty_print=pretty_print)
        outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
    else:
        outfile.write('/>%s' % (eol_, ))
DirectPositionType.export=_export


# make attribute gml:id being printed with namespace
def _exportAttributes(self, outfile, level, already_processed, namespace_='gml:', name_='AbstractGMLType'):
    if self.id is not None and 'id' not in already_processed:
        already_processed.add('id')
        #outfile.write(' id=%s' % (self.gds_format_string(quote_attrib(self.id).encode(ExternalEncoding), input_name='id'), ))
        outfile.write(' %sid=%s' % (namespace_, self.gds_format_string(quote_attrib(self.id).encode(ExternalEncoding), input_name='id'), ))
    if self.extensiontype_ is not None and 'xsi:type' not in already_processed:
        already_processed.add('xsi:type')
        outfile.write(' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"')
        outfile.write(' xsi:type="%s"' % self.extensiontype_)

AbstractGMLType.exportAttributes = _exportAttributes



"""Moule atmos_phys.convert

Unit conversions for atmospheric measuremensts.
Sub modules for different domains (aerosol, gases, ...; see there)

Example:
    A typical sequence for converting data::


        from atmos_phys.convert.gas import ConvertEthanal
        cvt = ConvertEthanal('pmol/mol', 'g/m3')
        data = [1.023, 0.998, 1.11]
        cvt.convert_data(data)
        print(data)
        [2.011e-09, 1.962e-09, 2.182e-09]
        print(cvt.conversion_string())
        from 'pmol/mol' to 'g/m3' at 273.15 K, 1013.25 hPa, conversion factor 0.00000000196544
"""



from .generated.ebas_diffexport_generated import *
from .generated.ebas_diffexport_generated import quote_attrib, __all__

# force datetime in attribute export (dbState and diffState)
def exportAttributes(self, outfile, level, already_processed,
                     namespace_='diffexport:', name_='diffExportType'):
    if self.dbState is not None and 'dbState' not in already_processed:
        already_processed.add('dbState')
        outfile.write(' dbState=%s' % (quote_attrib(self.gds_format_datetime(
            self.dbState, input_name='dbState')), ))
    if self.diffState is not None and 'diffState' not in already_processed:
        already_processed.add('diffState')
        outfile.write(' diffState=%s' % (quote_attrib(self.gds_format_datetime(
            self.diffState, input_name='dbState')), ))
diffExportType.exportAttributes = exportAttributes

# avoid namespacedef_ in child element's export calls
def exportChildren(self, outfile, level, namespace_='diffexport:',
                   name_='diffExportType', fromsubclass_=False,
                   pretty_print=True):
    for deleted_ in self.deleted:
        deleted_.export(outfile, level, namespace_, name_='deleted',
                        pretty_print=pretty_print, namespacedef_='')
    for added_ in self.added:
        added_.export(outfile, level, namespace_, name_='added',
                      pretty_print=pretty_print, namespacedef_='')
diffExportType.exportChildren = exportChildren



# force datetime in attribute export (startTime and endTime)
def dit_exportAttributes(self, outfile, level, already_processed,
                         namespace_='diffexport:', name_='dataIntervalType'):
    if self.datasetKey is not None and 'datasetKey' not in already_processed:
        already_processed.add('datasetKey')
        outfile.write(' datasetKey="%s"' % self.gds_format_integer(
            self.datasetKey, input_name='datasetKey'))
    if self.startTime is not None and 'startTime' not in already_processed:
        already_processed.add('startTime')
        outfile.write(' startTime=%s' % (quote_attrib(self.gds_format_datetime(
            self.startTime, input_name="startTime")),))
    if self.endTime is not None and 'endTime' not in already_processed:
        already_processed.add('endTime')
        outfile.write(' endTime=%s' % (quote_attrib(self.gds_format_datetime(
            self.endTime, input_name="endTime")),))
dataIntervalType.exportAttributes = dit_exportAttributes




"""
NasaAmes1001/__init__.py
$Id: __init__.py 1167 2016-01-19 09:57:07Z pe $

NASA Ames 1001 module

History:
V.1.0.0  2013-04-15  pe  initial version

"""

from .base import NasaAmes1001Error
from .read import NasaAmes1001Read
from .write import NasaAmes1001Write
from nilutility.datatypes import HexInt

class NasaAmes1001(NasaAmes1001Read, NasaAmes1001Write):
    """
    Partial class NasaAmes1001: Put together the parts to the final class.
    """
    pass



"""
EBAS submission templates and template checks
"""

from .ozone import EbasTemplateOzone
from .nox import EbasTemplateNOxLev0, EbasTemplateNOxLev1
from .base import NoTemplate


ALL_TEMPLATES = [
    EbasTemplateOzone,
    EbasTemplateNOxLev0,
    EbasTemplateNOxLev1,
]


def get_template(fil, ignore_templatecheck=False):
    """
    Find the matching template definition for a given file object/

    Parameters:
        fil   ebas-io file object
    Returns:
        EbasTemplateXX one of the template objects
    Raises:
        NoTemplate when no matching template could be found
        RuntimeError on any fatal condition (e.g. multiple templates match)
    """
    ret = []
    for template in ALL_TEMPLATES:
        try:
            ret.append(template(fil, ignore_templatecheck=ignore_templatecheck))
        except NoTemplate:
            pass
    if not ret:
        raise NoTemplate()
    if len(ret) > 1:
        raise RuntimeError('Multiple matches: {}'.format(', '.join(
            [x.__class__.__name__ for x in ret])))
    return ret[0]



"""
ebas/domain/masterdata/dc.py
$Id: dc.py 2669 2021-06-28 15:01:46Z pe $

EBAS Base Class for Dataset Characteristics

This module implements the class DCBase.

History:
V.1.0.0  2013-11-27  pe  initial version

"""

from .offline_masterdata import CTOfflineMasterData
from ..base import EbasDomainError
from ebas.domain.masterdata.ip import EbasMasterIP
from ebas.domain.basic_domain_logic.datalevel import base_datalevel
from .co import PTR_MS_COMP

class DCError(EbasDomainError):
    """
    Exception class raised for invalid DC.
    """
    pass

class DCWarning(Exception):
    """
    Exception class raised for DC warnings.
    """
    def __init__(self, msg, ret=None):
        """
        Exception class used for warnings.
        Can hold a return value (used for warnings raised in DCBase).
        """
        Exception.__init__(self, msg)
        self.ret = ret


class DCListBase(CTOfflineMasterData):
    """
    Base class for handling a list of Dataset Characteristics.
    Used by EbasIO (without db context) and
    EbasDomDCList (with db context) - client classes

    A client class must feature following functionality:
     - list type (implement __item__ and __iter__ and append)
     - provide a class atrribute CLIENT_CLASS_ELEMENTS
    """
    # Needs to be set by client class.
    CLIENT_CLASS_ELEMENTS = None
    def __getitem__(self, name):
        """
        Needs to be implemented by client class.
        """
        raise NotImplementedError
    def __iter__(self):
        """
        Needs to be implemented by client class.
        """
        raise NotImplementedError
    def append(self, elem):
        """
        Needs to be implemented by client class.
        """
        raise NotImplementedError

    def sorted(self):
        """
        Sorts list of dc objects according to metadata sort order.
        Parameters:
            None
        Returns:
            generator: sorted list of objects
        """
        for ret in self.sorted_order():
            yield ret[1]

    def sorted_order(self):
        """
        Sorts list of dc objects according to metadata sort order and yields
        sortorder number and list element as tuples.
        Parameters:
            dc_list   list of dataset characteristics
        Returns:
            generator: sorted list of tuples (ordernumber, object)
        """
        return sorted([(self.META['SORTORDER'].index(k.CT_TYPE) \
                           if k.CT_TYPE in self.META['SORTORDER'] \
                            else 999,
                        k) for k in self])

    def as_str(self):
        """
        Generates a single string from a sorted list of DC.
        Parameters:
            None
        Returns:
            str comma separated, tag=value
        """
        return ", ".join(["{}={}".format(dc_.CT_TYPE, dc_.value_string())
                          for dc_ in self.sorted()])

    def as_dict(self):
        """
        Generates a single dictionary from a list of DC.
        The dictionary is constructed as CT_TYPE: CT_VAL_xxx pairs
        Parameters:
            list of dc objects
        Returns:
            dictionary
        """
        ret = {}
        for dc_ in self:
            if dc_.CT_DATATYPE == 'CHR':
                ret[dc_.CT_TYPE] = dc_.DC_VAL_CHR
            elif dc_.CT_DATATYPE == 'INT':
                ret[dc_.CT_TYPE] = dc_.DC_VAL_INT
            else:
                ret[dc_.CT_TYPE] = dc_.DC_VAL_DBL
        return ret

    def dc_by_ct_type(self, ct_type):
        """
        Get a dc objetc from the list by CT_TYPE
        Parameters:
            ct_type    ct_type for lookup
        Returns:
            DCBase object or None
        """
        for dc_ in self:
            if dc_.CT_TYPE == ct_type:
                return dc_
        return None

    def add_parse(self, ct_type, value, ft_type, co_comp_name, data_level=None):
        """
        Adds a new object by parsing a DC tag=value pair.
        Calls element classes parse method.
        Used for parsing and validity check without db connection (e.g. in
        nasa ames parsing)
        Parameters:
            ct_type       characteristics type (tag name)
            value         characteristics value, string including unit if needed
            ft_type       ft_type of the dataset
                          used for validity checks
            co_comp_name  compornent name of the dataset
                          used for validity checks
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            None
        Raises:
            DCError, DCWarning
        """
        try:
            charac = self.__class__.CLIENT_CLASS_ELEMENTS.parse(
                ct_type, value, ft_type, co_comp_name, data_level=data_level)
        except DCWarning as excpt:
            if excpt.ret:
                self.append(excpt.ret)
            raise excpt

        # check if characteristic exists already:
        exist = 0
        for old in self:
            if old.CT_TYPE == charac.CT_TYPE:
                if old == charac and exist != 0:
                    exist = -1
                else:
                    exist = 1
        if exist == -1:
            self.append(charac)
            raise DCWarning(
                "multiple identical definitions for {}".format(ct_type))
        if exist == 1:
            raise DCError(
                "multiple different definitions for {}".format(ct_type))
        self.append(charac)

    def add(self, ct_type, value, ft_type, co_comp_name, data_level=None):
        """
        Add a new object to the list.
        Parameters:
            ct_type       characteristics type (tag name)
            value         characteristics value, string including unit if needed
            ft_type       ft_type of the dataset
                          used for validity checks
            co_comp_name  compornent name of the dataset
                          used for validity checks
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            None
        Raises:
            DCError, DCWarning
        """
        charac = self.__class__.CLIENT_CLASS_ELEMENTS(
            self.__class__.CLIENT_CLASS_ELEMENTS.setup_dict(
                ct_type, value, ft_type, co_comp_name, data_level=data_level))
        charac.__dict__['data_level'] = data_level
        charac.check()
        # check if characteristic exists already:
        exist = 0
        for old in self:
            if old.CT_TYPE == ct_type:
                if old == charac and exist != 0:
                    exist = -1
                else:
                    exist = 1
        if exist == -1:
            raise DCWarning(
                "multiple identical definitions for {}".format(ct_type))
        if exist == 1:
            raise DCError(
                "multiple different definitions for {}".format(ct_type))
        self.append(charac)

    def delete_dc(self, ct_type):
        """
        Remove a characteristic from the list
        Parameters:
            ct_type       characteristics type (tag name)
        Returns:
            None
        """
        for i, char in reversed(list(enumerate(self))):
            if char.CT_TYPE == ct_type:
                del self[i]

class DCBase(CTOfflineMasterData):
    """
    Base class for handling Dataset Characteristics.
    Used by nasa_ames.dataset_characteristic (without db context) and
    ebas.domain.dc (with db context) - client classes

    A client class must feature following functionality:
     - attribute access to data elements e.g. obj.CT_TYPE == __getattr__
     - key access to data elements e.g. obj['DC_VAL_INT'] == __item__
     - iteration over keys == __iter__
    All tree are implemented with classes based on DataObject (used in
    ebas.nasa_ames) and classes based on EbasDomainDictBase (used in
    ebas.domain.dc)
    """

    # some hardcoded exceptional components (lev 0 exceptions):
    CT_EXCEPTIONAL = {
        "0": {
            # aps lev0:
            'Size bin side scatter': {
                'CT_TYPE': 'Size bin side scatter',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            'Size bin aerodynamic': {
                'CT_TYPE': 'Size bin aerodynamic',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            # dmps lev 0:
            'Size bin': {
                'CT_TYPE': 'Size bin',
                'CT_DATATYPE': 'INT',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            # ccnc lev 0:
            'Actual/target': {
                'CT_TYPE': 'Actual/target',
                'CT_DATATYPE': 'CHR',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            # cpc lev 0:
            # online_gc lev 0:
            'Nominal/measured': {
                'CT_TYPE': 'Nominal/measured',
                'CT_DATATYPE': 'CHR',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            # cloud_light_scattering_photometer:
            'Channel': {
                'CT_TYPE': 'Channel',
                'CT_DATATYPE': 'INT',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            # PTR-MS lev0
            'k': {
                'CT_TYPE': 'k',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 20,
                'CT_UNIT': None,
                },
            'XR': {
                'CT_TYPE': 'XR',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 21,
                'CT_UNIT': None,
                },
            'accuracy': {
                'CT_TYPE': 'accuracy',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 21,
                'CT_UNIT': '%',
                },
            'dwell_time': {
                'CT_TYPE': 'dwell_time',
                'CT_DATATYPE': 'DBL',
                'CT_SEQ': 22,
                'CT_UNIT': 's',
                },
            'background_method': {
                'CT_TYPE': 'background_method',
                'CT_DATATYPE': 'CHR',
                'CT_SEQ': 23,
                'CT_UNIT': None,
                },
            'calibration_method': {
                'CT_TYPE': 'calibration_method',
                'CT_DATATYPE': 'CHR',
                'CT_SEQ': 24,
                'CT_UNIT': None,
                },
            },
        }
    CV_EXCEPTIONAL = {
        "0": [
            # ACSM lev 0:
            ('Reference component', 'aerosol_mass_spectrometer',
             'ionization_efficiency'),
            ('Reference component', 'aerosol_mass_spectrometer',
             'relative_ionization_efficiency'),
            ('Location', 'aerosol_mass_spectrometer',
             'relative_humidity'),
            ('Location', 'aerosol_mass_spectrometer',
             'electric_tension'),
            ('Location', 'aerosol_mass_spectrometer',
             'electric_current'),
            ('Location', 'aerosol_mass_spectrometer',
             'flow_rate'),
            ('Location', 'aerosol_mass_spectrometer',
             'frequency'),
            ('Location', 'aerosol_mass_spectrometer',
             'electric_power'),
            ('Status type', 'aerosol_mass_spectrometer',
             'status'),
            # aps lev0:
            ('Status type', 'aps', 'status'),
            ('Size bin', 'aps', 'particle counts'),
            ('Size bin side scatter', 'aps', 'particle counts'),
            ('Size bin aerodynamic', 'aps', 'particle counts'),
            ('Dmin', 'aps', 'particle counts'),
            ('Dmax', 'aps', 'particle counts'),
            ('Location', 'aps', 'flow_rate'),
            ('Location', 'aps', 'electric_current'),
            ('Location', 'aps', 'electric_tension'),
            ('Location', 'aps', 'digital_input'),
            # dmps lev 0:
            ('Size bin', 'dmps', 'particle_diameter'),
            ('Size bin', 'dmps', 'particle_number_concentration'),
            ('Size bin', 'smps', 'particle_diameter'),
            ('Size bin', 'smps', 'particle_number_concentration'),
            ('Size bin', 'v-smps', 'particle_diameter'),
            ('Size bin', 'v-smps', 'particle_number_concentration'),
            ('Wavelength', 'filter_absorption_photometer',
                'equivalent_black_carbon_loading'),
            # filter_absorption_photometer (AE31) lev0:
            ('Location', 'filter_absorption_photometer',
                'sensing_beam_signal'),
            ('Location', 'filter_absorption_photometer',
                'equivalent_black_carbon'),
            ('Location', 'filter_absorption_photometer',
                'attenuation_coefficient'),
            ('Wavelength', 'filter_absorption_photometer',
                'filter_loading_compensation_parameter'),
            # neph lev0
            ('Location', 'nephelometer', 'flow_rate'),
            ('Wavelength', 'nephelometer',
             'aerosol_light_scattering_coefficient_zero_measurement'),
            ('Wavelength', 'nephelometer',
             'aerosol_light_backscattering_coefficient_zero_measurement'),
            ('Wavelength', 'nephelometer',
             'aerosol_light_rayleighscattering_coefficient_zero_measurement'),
            # neph lev0: special case for humidified neph measurements (NOAA)
            # here, two nephs are used (one dry ref and one humidified)
            # the scattering/backscattering data get a Location element
            # containing RefNeph or WetNeph
            ('Location', 'nephelometer', 'aerosol_light_scattering_coefficient'),
            ('Location', 'nephelometer', 'aerosol_light_backscattering_coefficient'),
            # CCNC lev 0:
            ('Actual/target', 'CCNC', 'supersaturation'),
            ('Actual/target', 'CCNC', 'temperature_gradient'),
            ('Actual/target', 'CCNC', 'temperature'),
            ('Actual/target', 'CCNC', 'flow_rate'),
            ('Actual/target', 'CCNC', 'electric_tension'),
            ('Location', 'CCNC', 'flow_rate'),
            ('Location', 'CCNC', 'electric_current'),
            ('Location', 'CCNC', 'electric_tension'),
            ('Status type', 'CCNC', 'status'),
            # cpc lev0:
            ('Nominal/measured', 'cpc', 'flow_rate'),
            ('Location', 'cpc', 'pulse_width'),

            # DMPS-CCNC lev 0:
            ('Actual/target', 'DMPS-CCNC', 'supersaturation'),
            ('Actual/target', 'DMPS-CCNC', 'temperature_gradient'),
            ('Actual/target', 'DMPS-CCNC', 'temperature'),
            ('Actual/target', 'DMPS-CCNC', 'flow_rate'),
            ('Actual/target', 'DMPS-CCNC', 'electric_tension'),
            ('Location', 'DMPS-CCNC', 'flow_rate'),
            ('Location', 'DMPS-CCNC', 'electric_current'),
            ('Location', 'DMPS-CCNC', 'electric_tension'),
            ('Status type', 'DMPS-CCNC', 'status'),

            # NOx lev0
            ('Status type', 'chemiluminescence_photolytic', 'status'),
            ('Status type', 'chemiluminescence_molybdenum', 'status'),
            ('Status type', 'chemiluminescence_photometer', 'status'),

            # VOC lev0:
            ('Status type', 'online_gc', 'status'),
            # Additional exceptions (Nominal/measured) depend on IP table
            # (all components). See below.

            # PTR-MS lev0
            ('Status type', 'PTR-MS', 'status'),
            ('Location', 'PTR-MS', 'electric_tension'),

            # cloud_light_scattering_photometer
            ('Channel', 'cloud_light_scattering_photometer', 'electric_tension'),
        ]}

    # online_gc, add for all components:
    EMI = EbasMasterIP()
    for ip_ in EMI:
        if ip_['RE_REGIME_CODE'] == 'IMG' and ip_['FT_TYPE'] == 'online_gc':
            CV_EXCEPTIONAL["0"].append(
                ('Nominal/measured', 'online_gc', ip_['CO_COMP_NAME']))

    # PTR-MS, add for all components:
    EMI = EbasMasterIP()
    for comp in PTR_MS_COMP:
        CV_EXCEPTIONAL["0"].append(('k', 'PTR-MS', comp))
        CV_EXCEPTIONAL["0"].append(('XR', 'PTR-MS', comp))
        CV_EXCEPTIONAL["0"].append(('accuracy', 'PTR-MS', comp))
        CV_EXCEPTIONAL["0"].append(('dwell_time', 'PTR-MS', comp))
        CV_EXCEPTIONAL["0"].append(('background_method', 'PTR-MS', comp))
        CV_EXCEPTIONAL["0"].append(('calibration_method', 'PTR-MS', comp))


    # Problem: for checking validity of DC values without DB connection, we
    # need an offline store of valid CT and CV data.
    # read offline masterdata for DC (dataset characteristics methods)
    # Those are fallback values, will be read from database if possible.
    CTOfflineMasterData.read_pickle_file()

    # must be implemented by derived classes:
    def __getattr__(self, name):
        raise NotImplementedError
    def __getitem__(self, name):
        raise NotImplementedError
    def __iter__(self):
        raise NotImplementedError

    def __lt__(self, other):
        """
        Less then operator method.
        Needed in py3 for sorting: DCList.sorted_order()
        Parameters:
            other   compare object
        Returns:
            Bool
        """
        return self.tuple() < other.tuple()

    def __init__(self, dbh):
        """
        Read cached ct masterdata if dbh is provided.
        """
        CTOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    @classmethod
    def get_ct(cls, ct_type, data_level=None):
        """
        Get the correct CT metadata element (regular or exceptional (if
        data_level is specified)).
        Parameters:
            ct_type       characteristics type (tag name)
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            ct masterdata
        Raises:
            DCError if not found
        """
        data_level = base_datalevel(data_level)
        ct_ = None
        try:
            ct_ = cls.META['CHARACTERISTICS'][ct_type]
        except KeyError:
            try:
                if data_level is not None:
                    ct_ = cls.CT_EXCEPTIONAL[data_level][ct_type]
            except KeyError:
                pass
        if ct_ is None:
            raise DCError("unknown metadata element '{}'".format(ct_type))
            # wording needs to be 'metadata element' because misspelled
            # metadata elements will also end up here in characteristics
            # when reading a NasaAmes file
        return ct_

    @classmethod
    def validate_ct(cls, ct_type, data_level=None):
        """
        Check if ct_type is a characteristic at all.
        Parameters:
            ct_type       characteristics type (tag name)
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            None
        Raises:
            DCError if not valid
        """
        data_level = base_datalevel(data_level)
        if ct_type not in cls.META['CHARACTERISTICS']:
            if data_level is None or \
               data_level not in cls.CT_EXCEPTIONAL or \
               ct_type not in cls.CT_EXCEPTIONAL[data_level]:
                raise DCError("unknown metadata element '{}'".format(ct_type))

    @classmethod
    def validate_cv(cls, ct_type, ft_type, co_comp_name, data_level=None):
        """
        Check if characteristic is valid for ft_type and co_comp_name.
        Parameters:
            ct_type       characteristics type (tag name)
            ft_type       ft_type of the dataset
            co_comp_name  compornent name of the dataset
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            None
        Raises:
            DCError if not valid
        """
        data_level = base_datalevel(data_level)
        if (ct_type, ft_type, co_comp_name) not in cls.META['VALIDITY']:
            if data_level is None or \
               data_level not in cls.CV_EXCEPTIONAL or \
               (ct_type, ft_type, co_comp_name) not in \
               cls.CV_EXCEPTIONAL[data_level]:
                raise DCError(
                    "illegal metadata element '{}' for instrument '{}' and "
                    "component '{}'".format(ct_type, ft_type, co_comp_name))

    @classmethod
    def parse(cls, ct_type, value, ft_type, co_comp_name, data_level=None):
        # pylint: disable-msg=R0913
        # R0913: Too many arguments
        """
        Factory method. Creates a new object by parsing a DC tag=value pair.
        Check for validity.
        Used for parsing and validity check without db connection (e.g. in
        nasa ames parsing)
        Parameters:
            ct_type       characteristics type (tag name)
            value         characteristics value, string including unit if needed
            ft_type       ft_type of the dataset
                          used for validity checks
            co_comp_name  compornent name of the dataset
                          used for validity checks
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            new DatasetCharacteristic object
        Raises:
            DCError
        """
        ct_ = cls.get_ct(ct_type, data_level=data_level)
        unit = ct_['CT_UNIT']
        if unit is not None:
            # this value needs the unit appended
            if not value.endswith(unit):
                raise DCError('metadata value {}: missing unit ({})'.
                              format(ct_type, unit))
            value = value[:-len(unit)].strip()
        dc_dict = cls.setup_dict(ct_type, value, ft_type, co_comp_name,
                                 data_level=data_level)
        obj = cls(dc_dict)
        obj.__dict__['data_level'] = data_level
        dependencies = []
        if ft_type is None:
            dependencies.append("instrument type")
        if co_comp_name is None:
            dependencies.append("component name")
        if dependencies:
            raise DCWarning(
                'metadata element {} could not be checked for validity because '
                '{} is not specified correctly'
                .format(ct_type, ' and '.join(dependencies)), obj)
        else:
            obj.check()
        return obj

    @classmethod
    def setup_dict(cls, ct_type, value, ft_type, co_comp_name, data_level=None):
        # pylint: disable-msg=R0913
        # R0913: Too many arguments
        """
        Setup a correct data dictionary from base attributes.
        Used for both parsing (e.g. in nasa ames) and object creation for domain
        Parameters:
            ct_type       characteristics type (tag name)
            value         characteristics value (may still be string, but
                          without unit)
            ft_type       ft_type of the dataset
                          used for validity checks
            co_comp_name  compornent name of the dataset
                          used for validity checks
            data_level    data level, for exceptional DC (lev0 exceptions)
        Returns:
            new DatasetCharacteristic object
        Raises:
            DCError
        """
        dc_dict = \
            {
                'CT_TYPE': ct_type,
                'FT_TYPE': ft_type,
                'CO_COMP_NAME': co_comp_name
            }
        ct_ = cls.get_ct(ct_type, data_level=data_level)
        ct_datatype = ct_['CT_DATATYPE']
        dc_dict['CT_DATATYPE'] = ct_datatype
        if ct_datatype == 'INT':
            try:
                value = int(value)
            except ValueError:
                raise DCError('metadata value {}: must be type int'.
                              format(ct_type))
        if ct_datatype == 'DBL':
            try:
                value = float(value)
            except ValueError:
                raise DCError('metadata value {}: must be type float'.
                              format(ct_type))
        dc_dict['DC_VAL_' + str(ct_datatype)] = value
        return dc_dict

    def check(self):
        """
        Checks the object for validity.
        Raises:
            DCError
        """
        # get data_level, only is object has been set up with data_level
        # i.e. with parse(), else data_level is ignored and no exceptional
        # CT will be accepted
        data_level = None
        if 'data_level' in self.__dict__:
            data_level = self.data_level
        ct_ = self.get_ct(self.CT_TYPE, data_level=data_level)
        self.validate_cv(self.CT_TYPE, self.FT_TYPE, self.CO_COMP_NAME,
                         data_level=data_level)
        
        if ct_['CT_DATATYPE'] != self.CT_DATATYPE:
            raise DCError("metadata element '{}' must be {}, not {}".format(
                self.CT_TYPE, ct_['CT_DATATYPE'], self.CT_DATATYPE))
        # The following are internal coding errors: raise TypeError
        if (self.CT_DATATYPE == 'INT' and \
            ('DC_VAL_INT' not in self or self.DC_VAL_INT is None)) or \
           (self.CT_DATATYPE == 'DBL' and
            ('DC_VAL_DBL' not in self or self.DC_VAL_DBL is None)) or \
           (self.CT_DATATYPE == 'CHR' and
            ('DC_VAL_CHR' not in self or self.DC_VAL_CHR is None)):
            raise TypeError("metadata element '{}': {} ".
                            format(self.CT_TYPE, self.CT_DATATYPE) +
                            'value must be provided')
        if (self.CT_DATATYPE == 'INT' and \
            (('DC_VAL_DBL' in self and self.DC_VAL_DBL != None) or \
             ('DC_VAL_CHR' in self  and self.DC_VAL_CHR != None))) or \
           (self.CT_DATATYPE == 'DBL' and
            (('DC_VAL_INT' in self and self.DC_VAL_INT != None) or \
             ('DC_VAL_CHR' in self  and self.DC_VAL_CHR != None))) or \
           (self.CT_DATATYPE == 'CHR' and
            (('DC_VAL_INT' in self and self.DC_VAL_INT != None) or \
             ('DC_VAL_DBL' in self  and self.DC_VAL_DBL != None))):
            raise TypeError("metadata element '{}': ".format(self.CT_TYPE) +\
                          'multiple values with different datatypes')

    def tuple(self):
        """
        Generates a tag/value tuple.
        Parameters:
            None
        Returns:
            value tuple for a characteristic (<tag>, <value>[, <unit>])
        """
        # get data_level, only is object has been set up with data_level
        # i.e. with parse(), else data_level is ignored and no exceptional
        # CT will be accepted
        data_level = None
        if 'data_level' in self.__dict__:
            data_level = self.data_level
        ct_ = self.get_ct(self.CT_TYPE, data_level=data_level)
        return (self.CT_TYPE, self['DC_VAL_' + self.CT_DATATYPE], ct_['CT_UNIT'])

    @property
    def value(self):
        """
        Get the value for a DC.
        Parameters:
            None
        Returns:
            value
        """
        return self['DC_VAL_' + self.CT_DATATYPE]

    @property
    def unit(self):
        """
        Get the unit for a DC.
        Parameters:
            None
        Returns:
            unit
        """
        # get data_level, only is object has been set up with data_level
        # i.e. with parse(), else data_level is ignored and no exceptional
        # CT will be accepted
        data_level = None
        if 'data_level' in self.__dict__:
            data_level = self.data_level
        ct_ = self.get_ct(self.CT_TYPE, data_level=data_level)
        if ct_['CT_UNIT']:
            return ct_['CT_UNIT']
        return None

    def value_string(self):
        """
        Generates a value string ("<value> <unit>") that can be used for output.
        Parameters:
            None
        Returns:
            value string
        """
        ret = "{}".format(self.value)
        if self.unit is not None:
            ret += ' ' + self.unit
        return ret



"""
ebas/nasa_ames/parse_base.py
$Id: parse_base.py 1527 2017-02-15 18:36:36Z pe $

base class for parser for file input functionality for EBAS NASA Ames module

History:
V.1.0.0  2013-06-22  pe  initial version

"""
from ..base import EbasNasaAmesBase

class NasaAmesPartialReadParserBase(# pylint: disable=R0901, W0223,
        EbasNasaAmesBase):
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    # R0903: Too few public methods
    #     This is a base class.
    """
    Base class for Parser object.
    """
    pass



"""
$Id: eanet_filter.py 2749 2021-11-25 00:34:17Z pe $

Basic EANET Filter functionality is implemented in the EanetFilter class in
fileformats/EANET.
The ebas/io/eanet_filter (this) module is about mapping the EANET files to the
ebas IO data model.
"""

from .read_filter import EanetFilterPartialRead

class EbasEanetFilter(EanetFilterPartialRead):
    """
    EANET Filter I/O object.
    """




"""
nilutility/string_helper.py
$Id: string_helper.py 2454 2020-04-24 20:22:51Z pe $

Module for string helper functions.
This module is just a helper for dealing with common string manipulation tasks.
"""

import re
import csv
import six

class ListSplitterNewlineError(Exception):
    """
    Exception class for string parsing errors (list_splitter).
    """
    pass

class ListSplitterNullbyteError(Exception):
    """
    Exception class for string parsing errors (list_splitter).
    """
    pass

def _py2_unicode_csv_reader(unicode_csv_data, dialect=csv.excel, **kwargs):
    """
    The python2 csv module doesn't do Unicode; encode temporarily as UTF-8 and
    afterwards back to unicode.
    Parameters:
        unicode_csv_data   opened file or list of str
        dialect            csv dialect (see csv module)
        **kwargs           delimiter, skipinitialspace, ...
                           (see csv module)
    Returns:
        generator of lists (each list representing one line of csv)
    """
    csv_reader = csv.reader(_utf_8_encoder(unicode_csv_data),
                            dialect=dialect, **kwargs)
    for row in csv_reader:
        # decode UTF-8 back to Unicode, cell by cell:
        yield [cell.decode('utf-8') for cell in row]

def _utf_8_encoder(unicode_csv_data):
    """
    Helper for encoding unicode data to utf8.
    Parameters:
        unicode_csv_data    opened file or list of unicode strings
    Returns:
        utf-8 encoded text
    """
    for line in unicode_csv_data:
        yield line.encode('utf-8')


def list_splitter(string, delimiter=' ', skipinitialspace=True):
    """
    Splits a string. Double quoted parts are untouched.
    Parameters:
        string        string to be splitted
        delimiter     splitting character
        skipinitialspace
                      spaces after delimiter are ignored
    Returns:
        [part,...] splitted parts as list
    Raises:
        ListSplitterNewlineError
            if newline inside string
        csv.Error
            all other (unexpected/unknown error conditions)
    Synopsis: see test cases
    """
    try:
        if six.PY2:
            return next(_py2_unicode_csv_reader(
                [string], delimiter=delimiter,
                skipinitialspace=skipinitialspace))
        else:
            csv_reader = csv.reader(
                [string], dialect=csv.excel, delimiter=delimiter,
                skipinitialspace=skipinitialspace)
            for row in csv_reader:
                return [cell for cell in row]
    except csv.Error as expt:
        # seems to be an old issue: "newline inside string" seems to be
        # not thrown anymore in recent versions of the csv module.
        # however on ratatoskr this is currently a problem 
        # (status 2016-12-06, ubuntu 12.04, python 2.7.3)
        # this is an issue, we catch the _csv.Error exception here,
        # throw a ListSplitterNewlineError which is cought in the ebas.io parser.
        # TODO: may be deleted in th future
        # e.g. ubuntu 16.04, python 2.7.11, this is not an issue
        #      MacOS, python.org 2.7.10 also no issue
        if str(expt) == 'newline inside string':
            raise ListSplitterNewlineError(str(expt))
        if str(expt) == 'line contains NULL byte':
            raise ListSplitterNullbyteError(str(expt))
        raise expt

def list_joiner(lst, delimiter=' ', insert_space=False):
    """
    Joins a list to a string. If the delimiter occurs in an element of the
    list or any element has leading or trailing whitespaces, this element
    is escaped (double quoted).
    Parameters:
        lst       list to be joined
        delimiter delimiter character
    Returns:
        string
    Synopsis: see test cases
    """
    # make a copy of the list and
    # - substitute list elements None by empty string
    # - wrap elements containing a comma in double quotes
    lst = ['' if elem is None
           else '"' + re.sub('"', '""', elem) +'"' \
               if re.search(delimiter, elem) or re.search('"', elem)
           else elem
           for elem in lst]
    if insert_space:
        delimiter += ' '
    return delimiter.join(lst)




"""
nilu/datatypes.py
$Id: datatypes.py 2613 2021-02-18 10:58:49Z pe $

Module for generic datatypes.

History:
V.1.0.0  2012-10-05  pe  initial version

"""

import datetime
import six

class DataObject(dict):
    """
    Basic Object used for data dictionaries.
    This can be used as dictionary or in class attribute notation:

    # CREATION:
    dat=DataDictObj(bla=2, wtf=3)
    # same as:
    dat=DataDictObj({'bla'=2, 'wtf': 3})
    # same as:
    dat = DataDictObj([('bla', 2), ('wtf', 3)])

    # USAGE:
    dat.bla
    # same as:
    dat['bla']

    # SETTING:
    dat.bla = 2
    # same as:
    dat['bla'] = 2

    # EXCEPTIONS:
    dat.x
     --> AttributeError
    dat['x']
     --> KeyError
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as expt:
            raise AttributeError(str(expt))

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError as expt:
            raise AttributeError(expt)

    def copy(self):
        return DataObject(self)

def recursive_data_object(*args, **kwargs):
    """
    Factory for buildinfg a recursive DataObject instance.
    Every "child" which is a dict will become a DataObject (subclass of dict).
    Parameters:
        passed on to the dict instanciation
    Returns:
        DataObject
    """
    def _recurse(elem):
        """
        Recurse into other structures (dict, list, set, tuple)
        """
        if type(elem) == dict:
            # use type: we only convert "plain" dicts, not subclasses!
            return recursive_data_object(elem)
        elif isinstance(elem, dict):
            for key in elem.keys():
                elem[key] = _recurse(elem[key])
        if isinstance(elem, list):
            for i in range(len(elem)):
                elem[i] = _recurse(elem[i])
            return elem
        if isinstance(elem, set):
            ret = type(elem)()
            for member in elem:
                ret.add(_recurse(member))
            if ret == elem:
                return(elem)
            return(ret)
        if isinstance(elem, tuple):
            ret = []
            for member in elem:
                ret.append(_recurse(member))
            ret = type(elem)(ret)
            if ret == elem:
                return(elem)
            return(ret)
        return(elem)

    ret = DataObject(*args, **kwargs)
    for key in ret.keys():
        ret[key] = _recurse(ret[key])
    return ret

class Histogram(dict):
    """
    Class for creating histograms, based on dict.
    SYNOPSIS:
        # generate
        hist = Histogram()
        # collect:
        hist.increment('a')
        hist.increment('b')
        hist.increment('c')
        hist.increment('b')
        hist.increment('b')
        # results:
        res = hist.get_max()
        # res == (3, ['b'])
    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.unhashables = []

    def increment(self, value, inc=1):
        """
        Increments the histogram value.
        Parameters:
            value    histogram value
            inc      increment by (number)
        Returns:
            None
        """
        try:
            if not value in list(self.keys()):
                self[value] = inc
            else:
                self[value] += inc
        except TypeError:
            # if value is unhashable, use workaround
            for k in self.unhashables:
                if k[0] == value:
                    k[1] += 1
                    break
            else:
                self.unhashables.append([value, 1])

    def get_max(self):
        """
        Finds the maximum occorence.
        Parameters:
            None
        Returns:
            (maxocc, [keys])
        """
        if len(self) + len(self.unhashables) == 0:
            return (0, [])
        maxocc = max([self[x] for x in list(self.keys())] +\
                     [k[1] for k in self.unhashables])
        return (maxocc, [k for k in list(self.keys()) if self[k] == maxocc] +\
                        [k[0] for k in self.unhashables if k[1] == maxocc])


# Class HexInt

def returnthisclassfrom(methods, otype):
    """
    Decorator function for making certain methods of a class return an instance
    of the class itself instead of another type (all methods in list returning
    otype will return class instances).
    This works only if class can be instancianated from otype.
    """
    def gen_wrapper(cls, method):
        """
        Generate the customized wrapper function according to desired return
        type and method to be called.
        """
        def wrapper(*args, **kwargs):
            """
            Customized wrapper function for the original method.
            Will be created according to cls and method parameters of
            gen_wrapper.
            """
            res = method(*args, **kwargs)
            if isinstance(res, otype):
                return cls(res)
            return res
        return wrapper
    def wrap_methods(cls):
        """
        Wrap all relevant methods in the customized wrapper function.
        """
        for met in methods:
            method = getattr(cls, met)
            setattr(cls, met, gen_wrapper(cls, method))
        return cls
    return wrap_methods

# Compatibility issues python 2 and 3:
# __div__ is only implemented in python2. The python3 __truediff__ should
# of course _not_ be in the list.
if six.PY2:
    return_this_class_methods = (
        '__abs__', '__add__', '__and__', '__div__', '__floordiv__',
        '__invert__', '__lshift__', '__mod__', '__mul__', '__neg__', '__or__',
        '__pow__', '__rshift__', '__sub__', '__trunc__', '__xor__')
else:
    return_this_class_methods = (
        '__abs__', '__add__', '__and__', '__floordiv__',
        '__invert__', '__lshift__', '__mod__', '__mul__', '__neg__', '__or__',
        '__pow__', '__rshift__', '__sub__', '__trunc__', '__xor__')
@returnthisclassfrom(return_this_class_methods, int)
class HexInt(int):
    """
    Class for representing a hex integer value.
    Modifications from int:
        - interpret  initialization string as hex if it is a string
        - modify most special methods to return HexInt objects instead of int
        - repr and str return hex strings
    Synopsis:
        Initialization:
            HexInt(0xa), HexInt('a'), HexInt('0xa'), HexInt(10), HexInt(012)
            all do the same thing:
            (HexInt(0xa), HexInt(0xa), HexInt(0xa), HexInt(0xa), HexInt(0xa))

        repr(HexInt(0xa)) => 'HexInt(0xa)'

        str(HexInt(0xa)) => '0xa'

        HexInt(0xf) + 2 => HexInt(0x11) but 2 + HexInt(0xf) => 17 and
            HexInt(0xf) + 2.1 => 17.1

        HexInt(0xf) / 2 => HexInt(0x7) but 15 / HexInt(2) => 7 and
            HexInt(0xf) / 2.1 => 7.142857142857142

        HexInt(0xf) % 2 => HexInt(0x1) but 15 % HexInt(2) => 1 and
            HexInt(0xf) % 2.0 => 1.0

        divmod(HexInt(0xf),2) =>(HexInt(0x7), HexInt(0x1)) but
            divmod(15, HexInt(2)) => (7, 1) and
            divmod(HexInt(0xf), 2.0) => (7.0, 1.0)
    """
    def __new__(cls, *args):
        if len(args) == 1 and isinstance(args[0], six.string_types):
            obj = int.__new__(cls, args[0], 16)
        else:
            obj = int.__new__(cls, *args)
        return obj
    def __repr__(self):
        return 'HexInt({:#x})'.format(self)
    def __str__(self):
        return '{:#x}'.format(self)
    def __divmod__(self, other):
        return tuple(HexInt(x) if isinstance(x, int) else x
                     for x in divmod(int(self), other))

class datetimeprec(datetime.datetime): # pylint: disable=C0103
    # C0103: Invalid name (should match [A-Z_][a-zA-Z0-9]
    #   -> The original class was called datetime. A similar name is wanted.
    """
    Enhanced dateime.datetime class.
    + Adds one attribute for defining a precision with which the object was
    created. This attribute can either be specified explicitly on creation as
    a keyword argument, or will be set implicitly according to other parameters
    passed.
    + Allows creation of datetime object with year only or year and month only.
    """
    def __new__(cls, *args, **kwargs):
        """
        datetiome.datetime is immutable, so we must override __new__()
        """
        prec = None
        if 'prec' in kwargs:
            prec = kwargs['prec']
            cls.checkprec(prec)
            del kwargs['prec']
        if prec is None:
            if len(args) >= 7 or 'microsecond' in kwargs:
                prec = 'f'
            elif len(args) == 6 or 'second' in kwargs:
                prec = 'S'
            elif len(args) == 5 or 'minute' in kwargs:
                prec = 'M'
            elif len(args) == 4 or 'hour' in kwargs:
                prec = 'H'
            elif len(args) == 3 or 'day' in kwargs:
                prec = 'd'
            elif len(args) == 2 or 'month' in kwargs:
                prec = 'm'
            else:
                prec = 'Y'
        if len(args) < 2 and \
           [tag for tag in ['microsecond', 'second', 'minute', 'hour',
                            'day', 'month'] if tag in kwargs] == []:
            kwargs['month'] = 1
        if len(args) < 3 and \
           [tag for tag in ['microsecond', 'second', 'minute', 'hour',
                            'day'] if tag in kwargs] == []:
            kwargs['day'] = 1
        ret = datetime.datetime.__new__(cls, *args, **kwargs)
        ret.prec = prec
        return ret

    def __repr__(self):
        """
        Representation of a datetimeprec object.
        Parameters:
            None
        Returns:
            repr string
        """
        return "datetimeprec: " + self.strftime("%Y-%m-%dT%H:%M:%S.%f") + \
               ", prec=" + self.prec

    @staticmethod
    def checkprec(prec):
        """
        Checks the precision string for validity
        Parameters:
            prec   precision string
        """
        if not isinstance(prec, six.string_types):
            raise TypeError(
                "prec: string expected, not {}".format(prec.__class__.__name__))
        if prec not in ('Y', 'm', 'd', 'H', 'M', 'S', 'f'):
            raise ValueError("illigal precission string '{}'".format(prec))

    @classmethod
    def fromdatetime(cls, dto, prec):
        """
        Creates a datetimeprec object from a datetime.datetime object.
        Precision must be set, the original precision cannot be guessed from
        the datetime object.
        Parameters:
            dto     datetime.datetime object
            prec    precision
        """
        kwargs = {}
        cls.checkprec(prec)
        if prec in ('Y', 'm', 'd', 'H', 'M', 'S', 'f'):
            kwargs['year'] = dto.year
        if prec in ('m', 'd', 'H', 'M', 'S', 'f'):
            kwargs['month'] = dto.month
        if prec in ('d', 'H', 'M', 'S', 'f'):
            kwargs['day'] = dto.day
        if prec in ('H', 'M', 'S', 'f'):
            kwargs['hour'] = dto.hour
        if prec in ('M', 'S', 'f'):
            kwargs['minute'] = dto.minute
        if prec in ('S', 'f'):
            kwargs['second'] = dto.second
        if prec in ('f',):
            kwargs['microsecond'] = dto.microsecond
        kwargs['tzinfo'] = dto.tzinfo
        kwargs['prec'] = prec
        return cls.__new__(cls, **kwargs)

    def todatetime(self):
        """
        Converts into a normat datetime.
        Returns:
            datetime
        """
        return datetime.datetime(
            self.year, self.month, self.day,
            self.hour, self.minute, self.second, self.microsecond,
            tzinfo=self.tzinfo)

class ControlledDictList(list):
    """
    A versatile list of controlled dictionaries.
    Each element is a dict. The allowed keys of the dictionaries as well as the
    values for certain keys can be limited. elements can be added either as dict
    or as tuple/list.
    """
    def __init__(self, *args, **kwargs):
        """
        Set up the list. The allowed keys for the dictionaries is limited, the
        allowed values for each key can be limited.
        Parameters:
            keys       key names used for the elements of each dict (mandatory)
            <key-name> list of allowed values for the key
        Returns:
            none
        """
        self.allowed = {}
        if 'keys' in kwargs:
            self.keys = kwargs['keys']
            del kwargs['keys']
        else:
            raise ValueError('no keys defined')
        for key in self.keys:
            if key in kwargs:
                # allowed values for key:
                self.allowed[key] = kwargs[key]
                del kwargs[key]
        if kwargs:
            raise ValueError('unexpected keys: {}'.format(
                ', '.join(list(kwargs.keys()))))
        super(ControlledDictList, self).__init__(*args, **kwargs)

    def add(self, *args, **kwargs):
        """
        Adds one element (dict) to the list.
        Parameters:
            all keys defined by __init__
            or a list/tuple containing as many elements as self.keys
        """
        if args and kwargs:
            raise RuntimeError('either args or kwargs allowed, no mix')
        if args:
            if len(args) != len(self.keys):
                raise RuntimeError('number of args does not match number of '
                                   'keys')
            self.append({key: args[i] for i, key in enumerate(self.keys)})
        else:
            self.append(kwargs)

    def append(self, element):
        """
        Override append method. Additionally check for ControlledDictList logic.
        Parameters:
            element    element to append
        """
        # must be dict
        if not isinstance(element, dict):
            raise TypeError('only dict allowed')

        # only keys defined in self.keys are valid; all are mandatory
        err = [x for x in self.keys if x not in element]
        if err:
            raise RuntimeError("Missing keys: {}".format(', '.join(err)))
        err = [x for x in list(element.keys()) if x not in self.keys]
        if err:
            raise RuntimeError("Extra keys: {}".format(', '.join(err)))

        # check allowed values for certain keys
        for key in self.allowed:
            if element[key] not in self.allowed[key]:
                raise ValueError("'{}' is not allowed for key '{}'".format(
                    element[key], key))
        super(ControlledDictList, self).append(element)

    def insert(self, pos, element):
        """
        Override insert method. Additionally check for ControlledDictList logic.
        Parameters:
            pos        position to insert at
            element    element to insert
        """
        # must be dict
        if not isinstance(element, dict):
            raise TypeError('only dict allowed')

        # only keys defined in self.keys are valid; all are mandatory
        err = [x for x in self.keys if x not in element]
        if err:
            raise RuntimeError("Missing keys: {}".format(', '.join(err)))
        err = [x for x in list(element.keys()) if x not in self.keys]
        if err:
            raise RuntimeError("Extra keys: {}".format(', '.join(err)))

        # check allowed values for certain keys
        for key in self.allowed:
            if element[key] not in self.allowed[key]:
                raise ValueError("'{}' is not allowed for key '{}'".format(
                    element[key], key))
        super(ControlledDictList, self).insert(pos, element)

    def _check_compatibility(self, other):
        """
        Check compatibility of lists in order to add to one another.
        """
        if type(self) != type(other):  # pylint: disable=C0123
            # C0123: Using type() instead of isinstance() for a typecheck.
            # --> this type comparison cannot be substituted by isinstance
            raise TypeError(
                "Only a list of the same type can be added to a {}, not {}"
                .format(self.__class__.__name__, other.__class__.__name__))
        if self.keys != other.keys or self.allowed != other.allowed:
            raise TypeError("{} to add must have the same "
                            "configuration".format(self.__class__.__name__))

    def __add__(self, other):
        """
        Overrides the __add__ method. Check configuration and return a
        EbasDomainIssueList.
        Parameters:
            other    the other list to add
        """
        self._check_compatibility(other)
        return type(self)(list.__add__(self, other))

    def __iadd__(self, other):
        """
        Override the __iadd__ method. Check configuration.
        Parameters:
            other    the other list to iadd
        """
        self._check_compatibility(other)
        super(ControlledDictList, self).__iadd__(other)
        return self



#!/Library/Frameworks/Python.framework/Versions/2.7/bin/python
import sys
from . import ebas_diffexport
import datetime
import pytz

state = datetime.datetime.utcnow()
state = pytz.utc.localize(state)
diff = ebas_diffexport.diffExportType(
    dbState=state,
    diffState=datetime.datetime(2016, 1, 1, tzinfo=pytz.timezone('UTC')))

start = datetime.datetime(2015, 1, 1, tzinfo=pytz.timezone('UTC'))
end = datetime.datetime(2015, 2, 1, tzinfo=pytz.timezone('UTC'))
deleted = ebas_diffexport.dataIntervalType(datasetKey=1, startTime=start, endTime=end)
diff.add_deleted(deleted)

start = datetime.datetime(2015, 3, 1, tzinfo=pytz.timezone('UTC'))
end = datetime.datetime(2015, 4, 1, tzinfo=pytz.timezone('UTC'))
deleted = ebas_diffexport.dataIntervalType(datasetKey=2, startTime=start, endTime=end)
diff.add_deleted(deleted)

start = datetime.datetime(2015, 1, 1, tzinfo=pytz.timezone('UTC'))
end = datetime.datetime(2016, 1, 1, tzinfo=pytz.timezone('UTC'))
added = ebas_diffexport.dataIntervalType(datasetKey=1, startTime=start, endTime=end)
diff.add_added(added)

start = datetime.datetime(2015, 1, 1, tzinfo=pytz.timezone('UTC'))
end = datetime.datetime(2016, 1, 1, tzinfo=pytz.timezone('UTC'))
added = ebas_diffexport.dataIntervalType(datasetKey=2, startTime=start, endTime=end)
diff.add_added(added)

#position=ebas.Position(gml.PointType(id='pos1', srsName='urn:ogc:def:crs:EPSG:6.6:4979', srsDimension=3, pos=gml.DirectPositionType(valueOf_='1 2 3')))
#res.add_DataSet(ebas.DataSetType(123456, 'NO0001R', 'Birkenes', 'filterpack', 'sulfate', Matrix='pm10', Unit='mg/m3', Statistics='arithmetic mean', ResolutionCode='1d', DataLevel='2', Position=position, HeightAGL=ebas.HeightAGL('m', 11), Values=(1,2,3,4,5)))
#result = ebas.Result()
#result.add_ResultSet(res)

sys.stdout.write('<?xml version="1.0" ?>\n')
diff.export(sys.stdout, 0, name_="diffExport")



"""
ebas/io/netcdf/base.py
$Id$

EBAS NetCDF base class

"""

from ..basefile.file import EbasJoinedFile
from ..base import EBAS_IOFORMAT_NETCDF

class EbasNetcdfBase(EbasJoinedFile):  # pylint: disable=R0901, W0223,
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Base class for all EBAS netcdf classes.
    """

    # IOFORMAT = EBAS_IOFORMAT_NETCDF  # set in final object, as long as we have
    # two versions.... 

    def __init__(self, *args, **kwargs):
        """
        Class initialization.
        """
        EbasJoinedFile.__init__(self, *args, **kwargs)
        self.ncfile = None
        # self.strict_global depends on version 1 or two, see netcdf.py

    def gen_filename(self, createfiles=False, destdir=None):
        # pylint: disable-msg=W0221
        # W0221: Arguments number differs from overridden method
        """
        Generates a file name according to NILU file name convention.
        Use the base class functionality but set the file name extension.
        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
        Return:
            None
        """
        EbasJoinedFile.gen_filename(self, createfiles=createfiles,
                                    destdir=destdir, extension=".nc")



"""
ebas/io/nasa_ames/base.py
$Id: base.py 2519 2020-10-30 16:13:09Z pe $

EBAS NASA Ames base class

"""

from ..basefile.file import EbasFile
from ...base import EbasIOError
from ..base import EBAS_IOFORMAT_NASA_AMES

class EbasNasaAmesReadError(EbasIOError):
    """
    Exception class for NasaAmes read errors.
    """
    pass

class EbasNasaAmesBase(EbasFile):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    # This is also an abstract class, just does not raise NotImplementedError
    """
    Base class for all EBAS Nasa Ames classes.
    """
    # set exception for generic read method
    READ_EXCEPTION = EbasNasaAmesReadError
    IOFORMAT = EBAS_IOFORMAT_NASA_AMES

    def __init__(self, *args, **kwargs):
        """
        Class initioalization.
        """
        EbasFile.__init__(self, *args, **kwargs)
        # add one additional attribute:
        self.nasa1001 = None
        self.strict_global = False   # EBAS Nasa Ames does NOT use strict_global

    def gen_filename(self, createfiles=False, destdir=None):
        # pylint: disable-msg=W0221
        # W0221: Arguments number differs from overridden method
        """
        Generates a file name according to NILU file name convention.
        Use the base class functionality but set the file name extension.
        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
        Return:
            None
        """
        EbasFile.gen_filename(self, createfiles=createfiles,
                              destdir=destdir, extension=".nas")



"""
ebas/nasa_ames/fileset.py
$Id: fileset.py 2721 2021-10-22 23:02:49Z pe $

EBAS NASA Ames module

History:
V.1.0.0  2012-09-20  pe  initial version

"""

import logging
from nilutility.datetime_helper import DatetimeInterval
from ebas.domain.basic_domain_logic.unit_convert import UnitConvert, \
    NoConversion
from .xmlwrap import xmlwrap_header, xmlwrap_trailer
from ..file.base import EBAS_IOFORMAT_NASA_AMES, EBAS_IOFORMAT_CSV, \
    EBAS_IOFORMAT_XML, EBAS_IOFORMAT_NETCDF1, EBAS_IOFORMAT_NETCDF, \
    EBAS_IOFORMAT_OPENDAP, \
    EBAS_IOSTYLE_SINGLECOLUMN, EBAS_IOSTYLE_MULTICOLUMN, \
    EBAS_IOSTYLE_MULTICOLUMN_PRECIP, EBAS_IOSTYLE_KEEP, \
    FLAGS_ONE_OR_ALL


class NoDataDOI(Exception):
    pass


class EbasIOResultSet(object):
    # pylint: disable-msg=R0902
    #  R0902: Too many instance attributes
    """
    Set of results (NASA Ames, CSV, XML, netcdf).
    """
    def __init__(self, outformat=EBAS_IOFORMAT_NASA_AMES,
                 outstyle=EBAS_IOSTYLE_MULTICOLUMN, createfiles=False,
                 destdir=None, xmlwrap=False, precip_amount=False,
                 expand=False, long_timeseries=True, flags=FLAGS_ONE_OR_ALL,
                 metadata_options=0, access_log=None, indexdb=None,
                 diff_xml=None, gen_converted=False, noextract=False,
                 doi=None):
        """
        Set up resultset object.
        Parameters:
            outformat   the (file)format of the results (triggers which type of
                        result object will be generated
            outstyle    EBAS_IOSTYLE_SINGLECOLUMN, EBAS_IOSTYLE_MULTICOLUMN,
                        EBAS_IOSTYLE_MULTICOLUMN_PRECIP, EBAS_IOSTYLE_KEEP
            createfiles bool, whether actual files should be generated or output
                        should go to stdout
            destdir     destination directory for output files
            xmlwrap     bool, result should be wrapped in xml container
                        (for ACTRIS webservice)
            precip_amount
                        bool, add precipitation amout datasets for concentartion
                        measurements automatically
            expand      bool, expand multicolumn output (add all datasets from
                        the same instrument which fit in the file)
            long_timeseries
                        produce long time series, even when metadata change over
                        time. Variable metadata are only supported by some file
                        formats. Those who do not support variable metadata
                        just ignore the argument (join method does nothing).
            flags       flag output behaviour (FLAGS_ONE_OR_ALL, FLAGS_COMPRESS,
                        FLAGS_ALL, FLAGS_NONE
            metadata_options
                        triggers output of optional metadata, currently only
                        EBAS_IOMETADATA_OPTION_SETKEY
            access_log  AccessLog object for logging the data access
            indexdb     IndexDb object for creation of a file index database
            diff_xml    EbasDiffExportXML object for creation of a diffxml
                        document
            gen_converted
                        bool, trigger the unit conversion for selected
                        parameters on output (add a converted variable, e.g.
                        ozone in nmol/mol)
            noextract   bool, trigges the suppression of the actual extraction
                        (needed in a special case for the OPeNDAP server)
            doi         DOI URI; special case: This Object represents a 
                        state DOI. Consequences for filenames and metadata
                        elements doi, doi_list
        """
        # pylint: disable-msg=R0913
        #  R0913: Too many arguments
        
        self.logger = logging.getLogger('EbasIO')
        self.results = []
        self.result_num = 0   # number of results generated in total
        self.outformat = outformat
        # local import to avoid circular import
        if self.outformat == EBAS_IOFORMAT_NASA_AMES:
            from ..file.nasa_ames import EbasNasaAmes
            self.result_class = EbasNasaAmes
        elif self.outformat == EBAS_IOFORMAT_XML:
            from ..file.xml   import EbasXML
            self.result_class = EbasXML
        elif self.outformat == EBAS_IOFORMAT_CSV:
            from ..file.csv   import EbasCSV
            self.result_class = EbasCSV
        elif self.outformat == EBAS_IOFORMAT_NETCDF1:
            from ..file.netcdf import EbasNetcdf1
            self.result_class = EbasNetcdf1
        elif self.outformat == EBAS_IOFORMAT_NETCDF:
            from ..file.netcdf import EbasNetcdf
            self.result_class = EbasNetcdf
        elif self.outformat == EBAS_IOFORMAT_OPENDAP:
            from ..file.opendap import EbasOPeNDAP
            self.result_class = EbasOPeNDAP
        else:
            raise ValueError('Unknown output format')
        self.outstyle = outstyle
        self.createfiles = createfiles
        self.destdir = destdir
        self.xmlwrap = xmlwrap
        self.precip_amount = precip_amount
        self.expand = expand
        self.long_timeseries = long_timeseries
        self.prec_di_export = {}
        self.prec_var_nondomain = {}
        self.flags = flags
        self.metadata_options = metadata_options
        self.clear_caches = None
        self.access_log = access_log
        self.indexdb = indexdb
        self.diff_xml = diff_xml
        self.gen_converted = gen_converted
        self.noextract = noextract
        self.export_state = None
        self.export_filter = None
        self.doi = doi
        self.nodata = True
        self.restricted = False
        if not self.noextract and self.xmlwrap:
            xmlwrap_header()
        if not self.noextract and self.outformat == EBAS_IOFORMAT_XML and \
           not createfiles:
            print('<?xml version="1.0" encoding="utf-8"?>')
            print('<ebas:Result xmlns:ebas="http://ebas.nilu.no/EBAS">')
        self._id = -1

    @property
    def unique_id(self):
        """
        Get a unique identifier.
        """
        self._id += 1
        return self._id

    def add_datasets(self, ds_list, on_the_fly=True, clear_caches=None,
                     crit=None, time=None):
        """
        Adds datasets to the file set.

        Parameters:
            ds_list      iterator of dataset objects
            on_the_fly   export on the fly (creates output as soon as possible)
            clear_caches EbasDomainObjCache to be cleared when output is done
            crit         additional criteria to be checked for each export
                         snipplet (currently only project) (dictionary)
            time         time criteria, only used for extract_doi, else the
                         time criteria are taken from the dbctx
        Returns:
            None
        """
        prev_station = None
        self.clear_caches = clear_caches
        if self.export_state and self.export_state != ds_list.dbctx.state:
            raise RuntimeError(
                "only the same state can be added to a result set")
        self.export_state = ds_list.dbctx.state
        if self.export_filter and \
                self.export_filter != ds_list.dbctx.export_filter:
            raise RuntimeError(
                "only the same export filter can be added to a result set")
        self.export_filter = ds_list.dbctx.export_filter
        for dsobj in ds_list:
            # new station code: finalize previous results
            if on_the_fly:
                if prev_station and prev_station != dsobj.ST_STATION_CODE:
                    self.extract_all()
                prev_station = dsobj.ST_STATION_CODE
            nodata = True
            # loop through data intervals
            for di_ in dsobj.di:
                # find interval sniplets to be exported:
                # (changing projects, variable metadata or instrument metadata)
                intervals = di_.get_export_intervals(crit, time)
                for intv in intervals:
                    nodata = False
                    if intv[1]:
                        self.nodata = False
                        self.add_di(di_, intv[0])
                        if self.access_log and not self.noextract:
                            # suppress access_log when noextract is set
                            # the file object has to care for access_log
                            # if anything is output.
                            self.access_log.access_dataset(
                                di_.ds.DS_SETKEY,
                                DatetimeInterval(intv[0][0],intv[0][1]))
                        if self.diff_xml:
                            self.diff_xml.add_added(di_.ds.DS_SETKEY,
                                                    intv[0][0], intv[0][1])
                    else:
                        self.logger.warning(
                            "no access to setkey %d [%s, %s]",
                            dsobj.DS_SETKEY, intv[0][0], intv[0][1])
                        self.restricted = True
            if nodata:
                self.logger.warning(
                    "no data for setkey %d [%s, %s]",
                    dsobj.DS_SETKEY, dsobj.dbctx.time[0], dsobj.dbctx.time[1])
        # Results are possibly unsorted in the case where files are splited
        # horizontally (see below substituted in add_hdi)
        # We need to sort in order to make e.g. join_years work. 
        self.results = sorted(
            self.results,
            key=lambda x: \
               (x.sample_times[0][0],
                x.internal.fromdomain_tmp.di_obj[0].DS_SETKEY))


    def add_di(self, di_obj, intv):
        """
        Merge a data interval of one dataset into the existing file set.
        Depending on self.outformat it tries to merge the data into existing
        results, merges parts into existing results or adds a new file.
        Will be executed recursively if needed.
        Adds precipitation amount if needed/wanted.

        Parameters:
            di_obj   the di object to be added
            intv     interval to be merged in (start, end) (datetime.datetime)
        Returns:
            None
        """
        if not list(di_obj.su.sa.values(intv)):
            # no samples in interval, do not add data interval
            return
        if not self.result_class.LAZY_READ:
            # If we are not a LAZY_READ object (e.g. OPeNDAP), we prefer to
            # eagerly read here. We need to read the whole DI data anyway, and
            # fragmented output files suffered up to 30% performace decrease
            # due to lazy read.
            _ = list(di_obj.ts.values(intv))
        precip_dis = self._get_precip_dis(di_obj, intv)
        if self.outstyle in (EBAS_IOSTYLE_SINGLECOLUMN,
                             EBAS_IOSTYLE_MULTICOLUMN_PRECIP):
            # singlecolumn: always create a new file
            result = self.new_result()
            result.setup_from_domain(di_obj, intv, 0)
            for pdi in  precip_dis:
                if self.outstyle == EBAS_IOSTYLE_MULTICOLUMN_PRECIP:
                    prec_int = intv.overlap(pdi.VALIDTIME)
                    (success, rem_intv) = result.merge_var_from_domain(
                        pdi, prec_int, False, 2 if self.gen_converted else 0)
                    if success is False:
                        self._add_prec_di(
                            pdi,
                            DatetimeInterval(result.sample_times[0][0],
                                             result.sample_times[-1][1]))
                        if rem_intv:
                            raise RuntimeError(
                                'only expandable variables can force a file '
                                'split')
                else:
                    self._add_prec_di(
                        pdi,
                        DatetimeInterval(result.sample_times[0][0],
                                         result.sample_times[-1][1]))
            if self.gen_converted:
                try:
                    unit_convert = UnitConvert()
                    unit_convert.export_conv_params(
                        di_obj.ds.RE_REGIME_CODE, di_obj.ds.MA_MATRIX_NAME,
                        di_obj.ds.CO_COMP_NAME, di_obj.ds.pm.PM_UNIT,
                        di_obj.ds.DS_VOLUME_STD_PRESS,
                        di_obj.ds.DS_VOLUME_STD_TEMP)
                    # We do not care about the result here. We just know there
                    # is a possible conversion if there is no NoConversion
                    # exception.
                except NoConversion:
                    pass
                else:
                    result = self.new_result()
                    result.setup_from_domain(di_obj, intv, 1)
            return
        if self.outstyle != EBAS_IOSTYLE_MULTICOLUMN:
            raise RuntimeError('Illegal IOSTYLE for add_di')
        # multicolumn: try to merge interval into existing outputfiles
        for resnum, result in enumerate(self.results):
            (success, rem_intv) = result.merge_var_from_domain(
                di_obj, intv, True, 2 if self.gen_converted else 0)
            if success:
                for pdi in  precip_dis:
                    prec_int = intv.overlap(pdi.VALIDTIME)
                    (success, _) = result.merge_var_from_domain(
                        pdi, prec_int, False, 2 if self.gen_converted else 0)
                    if not success:
                        self._add_prec_di(pdi, prec_int)
                for di_, int_ in rem_intv:
                    # recursively add remaining intervals
                    self.add_di(di_, int_)
                return
            elif rem_intv:
                # success is False and rem_intv: we need to substitute this file
                del self.results[resnum]
                for di_, int_ in rem_intv:
                    # recursively add remaining intervals
                    self.add_di(di_, int_)
                return
            # else no merge, keep iterating

        # multicolumn: only if interval could not be merged - create new file:
        result = self.new_result()
        result.setup_from_domain(di_obj, intv, 2 if self.gen_converted else 0)
        for pdi in  precip_dis:
            prec_int = intv.overlap(pdi.VALIDTIME)
            (success, _) = result.merge_var_from_domain(
                pdi, prec_int, False, 2 if self.gen_converted else 0)
            if not success:
                self._add_prec_di(pdi, prec_int)

    def add_ebasfile(self, ebasfile):
        """
        Add an existing ebas file to the fileset.
        According to IOSTYLE, the file ist either added whole, or all variables
        are added to the fileset (in order to support single/mulicolumn etc)
        Parameters:
            ebasfile     an ebas.io file object
        Returns
            None
        """
        if self.outstyle == EBAS_IOSTYLE_KEEP:
            # no changes in single/multicolumn, just copy the file to an export
            # file
            outputfile = self.new_result()
            outputfile.from_ebasfile(ebasfile)
        else:
            precip_vars = []
            for i, _ in enumerate(ebasfile.variables):
                if ebasfile.is_precip_amount(i):
                    metadata = ebasfile.get_all_meta_for_var(i)
                    precip_vars.append({
                        'id': self.unique_id,
                        'metadata': metadata,
                        'sample_times': ebasfile.sample_times,
                        'values': ebasfile.variables[i].values_,
                        'flags': ebasfile.variables[i].flags,
                        })
            precip_vars_used = False
            for i, _ in enumerate(ebasfile.variables):
                if not ebasfile.is_precip_amount(i):
                    metadata = ebasfile.get_all_meta_for_var(i)
                    if ebasfile.is_precip_concentration(i):
                        precip_vars_used = True
                        use_precip = precip_vars
                    else:
                        use_precip = None
                    self.add_var_non_domain(
                        self.unique_id,
                        metadata,
                        ebasfile.sample_times,
                        ebasfile.variables[i].values_,
                        ebasfile.variables[i].flags,
                        is_hex=ebasfile.variables[i].is_hex,
                        precip_vars=use_precip)
            if not precip_vars_used:
                for precip in precip_vars:
                    self.add_var_non_domain(
                        precip['id'],
                        precip['metadata'],
                        precip['sample_times'],
                        precip['values'],
                        precip['flags'])

    def add_var_non_domain(
            self, id_, metadata, sample_times, values, flags,
            is_hex=False, precip_vars=None):
        """
        Adds a variable (source not from doamin) to the file set.

        Parameters:
            metadata      metadata dict
            sample_times  sample times list
            values        variable's data
            precip_vars   list of precip variables, each dict:
                          {id_, metadata, sample_times, values, flags}
        Returns:
            None
        """
        if precip_vars is None:
            precip_vars = []
        if len(sample_times) != len(values) or len(sample_times) != len(flags):
            raise RuntimeError("different sample times and values/flags")
        if not sample_times:
            # no samples in interval, do not add data interval
            return
        if self.outstyle in (EBAS_IOSTYLE_SINGLECOLUMN,
                             EBAS_IOSTYLE_MULTICOLUMN_PRECIP):
            # singlecolumn: always create a new file
            result = self.new_result()
            result.setup_non_domain()
            result.sample_times = list(sample_times)
            result._add_variable_nondomain(
                id_, metadata, values, flags,
                2 if self.gen_converted else 0,
                is_hex=is_hex)
            for prec in  precip_vars:
                if len(prec['values']) != len(sample_times) or \
                        len(prec['flags']) != len(sample_times):
                    raise RuntimeError(
                        "different sample times and precip values/flags")
                if self.outstyle == EBAS_IOSTYLE_MULTICOLUMN_PRECIP:
                    (success, _) = result.merge_var_non_domain(
                        prec['id'], prec['metadata'], sample_times,
                        prec['values'], prec['flags'],
                        2 if self.gen_converted else 0)
                    if not success:
                        self._add_prec_var_nondomain(prec)
                else:
                    self._add_prec_var_nondomain(prec)
            return

        if self.outstyle != EBAS_IOSTYLE_MULTICOLUMN:
            raise RuntimeError('Illegal IOSTYLE for add_di')

        # multicolumn: try to merge interval into existing outputfiles
        for result in self.results:
            (success, rem_intv) = result.merge_var_non_domain(
                id_, metadata, sample_times, values, flags,
                2 if self.gen_converted else 0,
                is_hex=is_hex)
            if success:
                for precip in precip_vars:
                    (success, _) = result.merge_var_non_domain(
                        precip['id'], precip['metadata'],
                        precip['sample_times'], precip['values'],
                        precip['flags'],
                        2 if self.gen_converted else 0)
                    if not success:
                        self._add_prec_var_nondomain(precip)
                for intv in rem_intv:
                    # recursively add remaining intervals
                    self.add_var_non_domain(
                        id_, metadata, intv[0], intv[1], intv[2],
                        is_hex=is_hex)
                return

        # multicolumn: only if interval could not be merged - create new file:
        result = self.new_result()
        result.setup_non_domain()
        result.sample_times = list(sample_times)
        result._add_variable_nondomain(
            id_, metadata, values, flags,
            2 if self.gen_converted else 0,
            is_hex=is_hex)
        for prec in  precip_vars:
            if len(prec['values']) != len(sample_times) or \
                    len(prec['flags']) != len(sample_times):
                raise RuntimeError(
                    "different sample times and precip values/flags")
            (success, _) = result.merge_var_non_domain(
                prec['id'], prec['metadata'], sample_times, prec['values'],
                prec['flags'],
                2 if self.gen_converted else 0,
                is_hex=False)
            if not success:
                self._add_prec_var_nondomain(prec)

    def _get_precip_dis(self, di_obj, intv):
        """
        Get the list of relevant precipitation amount di's if needed.
        Parameters:
            di_obj   the di object to be added
            intv     interval to be merged in (start, end) (datetime.datetime)
        Returns:
            precip_dis  (list of EbasDomDI relevant for export
        """
        if self.precip_amount and di_obj.ds.is_precip_concentration():
            precip_dis = list(di_obj.precip())
            i = 0
            while i < len(precip_dis):
                # overlap concentration  and amount time coverage, sample start
                # end times according to precip submission:
                prec_intv = precip_dis[i].su.sa.interval(
                    intv.overlap(precip_dis[i].VALIDTIME))
                if prec_intv is None:
                    # precip di has no coverage for interval
                    del precip_dis[i]
                elif precip_dis[i].have_access(prec_intv) == False:
                    self.logger.warning(
                        'no access to amount DS {} [{:19.19s}  {:19.19s}] '
                        'for concentration DS {}'.\
                            format(precip_dis[i].ds.DS_SETKEY,
                                   prec_intv[0].strftime('%Y-%m-%dT%H:%M:%S'),
                                   prec_intv[1].strftime('%Y-%m-%dT%H:%M:%S'),
                                   di_obj.ds.DS_SETKEY))
                    del precip_dis[i]
                else:
                    i += 1
            if not precip_dis:
                self.logger.warning(
                    'no precipitation amount for DS {} [{:19.19s}  {:19.19s}]'.\
                    format(di_obj.ds.DS_SETKEY,
                           intv[0].strftime('%Y-%m-%dT%H:%M:%S'),
                           intv[1].strftime('%Y-%m-%dT%H:%M:%S')))
        else:
            precip_dis = []
        return precip_dis

    def new_result(self):
        """
        Adds a new result to the resultset.

        Parameters:
            None
        Returns:
            result object (nasa_ames, xml, csv, netcdf)
        """
        if self.outformat == EBAS_IOFORMAT_OPENDAP:
            if self.noextract:
                # when noextract is set, access_log will be added in
                # make_dataset
                new_result = self.result_class(access_log=self.access_log,
                                               indexdb=self.indexdb)
            else:
                new_result = self.result_class(indexdb=self.indexdb)
        elif self.outformat == EBAS_IOFORMAT_XML:
            new_result = self.result_class(self.result_num,
                                           indexdb=self.indexdb)
        else:
            # EBAS_IOFORMAT_NASA_AMES, EBAS_IOFORMAT_CSV,
            # EBAS_IOFORMAT_NETCDF1, EBAS_IOFORMAT_NETCDF
            new_result = self.result_class(indexdb=self.indexdb)

        # This is important:
        # Whenever data are exported from domain to ebas.io, we set
        # the export_state and the export_filter in the output file's metadata.
        # Thus we make sure this information comes whis all files exported from
        # ebas.
        new_result.metadata.export_state = self.export_state
        new_result.metadata.export_filter = self.export_filter
        self.results.append(new_result)
        self.result_num += 1
        return new_result

    def extract_all(self):
        """
        Extracts all results in the set.
        """
        if not self.noextract:
            # first, expand all results if needed
            if self.expand and \
               self.outstyle == EBAS_IOSTYLE_MULTICOLUMN:
                for result in self.results:
                    result.expand_multicolumn()
            # then create additional precipitation results - those should be
            # single column and __not__ expanded.
            self._export_prec_di()
            self._export_prec_var_nondomain()
            if self.long_timeseries:
                self._join_years()
                if self.doi and len(self.results) == 0:
                    if self.nodata and self.restricted:
                        raise NoDataDOI(
                            'No data for DOI because of access restrictions')
                    raise RuntimeError('unexpectedly no data')
                if self.doi and self.outformat == EBAS_IOFORMAT_NETCDF and \
                        len(self.results) != 1:
                    # state DOI as NetCDF file should always go to one file...
                    # except: when for some period some ds have access
                    # restrictions. Then the files have different variables
                    # over the years.
                    # Fix here: take the longest sequence of data (#samples)
                    # and throw the rest. self.restricted is already set.
                    if self.restricted:
                        keep = max(enumerate([len(res.sample_times)
                                              for res in self.results]),
                                   key=lambda x: x[1])[0]
                        for i in reversed(range(len(self.results))):
                            if i != keep:
                                del  self.results[i]
                    else:
                        raise RuntimeError('DOI should have 1 NetCDF file')
            if self.doi:
                self.results[0].metadata.doi = [self.doi, None]
                if len(self.results) > 1:
                    i = 1
                    for result in self.results:
                        result.metadata.doi = [self.doi, i]
                        i += 1
            # now write all results
            for result in self.results:
                result.write(createfiles=self.createfiles, destdir=self.destdir,
                             xmlwrap=self.xmlwrap, flags=self.flags,
                             metadata_options=self.metadata_options)
        self.results = []
        if self.clear_caches:
            self.clear_caches.clear()

    def _join_years(self):
        """
        Join results according to the outputformat.
        """
        i = 0
        while i < len(self.results):
            j = i + 1
            while j < len(self.results):
                joined = self.results[i].join(self.results[j])
                if joined:
                    self.results[i] = joined
                    del self.results[j]
                else:
                    j += 1
            i += 1

    def close(self):
        """
        Close resultset: extract all remaining results and shut down.
        """
        self.extract_all()
        if not self.noextract and self.outformat == EBAS_IOFORMAT_XML and \
           not self.createfiles:
            print('</ebas:Result>')
        if not self.noextract and self.xmlwrap:
            xmlwrap_trailer()

    def _add_prec_di(self, pdi, time):
        """
        Add precipitation di and interval to the list of datasets that need
        to be extracted as single column results in the end.
        Parameters:
            pdi       di object of the precipitation amount
            time      time interval needed - tuple (start, end)
                      The time interval that is covered by the concentration
                      timeseries. (interval will be overlapped with amount di
                      coverage)
        Returns:
            None
        """
        # overlap concentration  and amount time coverage, sample start end
        # times according to precip submission:
        overlap = pdi.su.sa.interval(time.overlap(pdi.VALIDTIME))
        # add the overlapping interval
        if pdi in self.prec_di_export:
            self.prec_di_export[pdi].append(overlap)
        else:
            self.prec_di_export[pdi] = [overlap]


    def _export_prec_di(self):
        """
        Extracts the needed precipitation amount intervals that could not be
        added to a multicolumn file. Export is single column here.
        Parameters:
            None
        Returns:
            None
        """
        for di_ in self.prec_di_export:
            i = 0
            while i < len(self.prec_di_export[di_]):
                intv = self.prec_di_export[di_]
                j = i+1
                while j < len(intv):
                    if intv[i][0] < intv[j][1] and intv[i][1] > intv[j][0]:
                        # intervals overlap, join them!
                        intv[i][0] = min(intv[i][0], intv[j][0])
                        intv[i][1] = max(intv[i][1], intv[j][1])
                        del intv[j]
                        break
                    else:
                        j += 1
                else:
                    # if nothing changed in j loop, export and continue with
                    # next i
                    new = self.new_result()
                    new.setup_from_domain(di_, self.prec_di_export[di_][i], 0)
                    i += 1
        self.prec_di_export = {}

    def _add_prec_var_nondomain(self, prec):
        """
        Add precipitation precipitation variable to the list of variables that
        need to be extracted as single column results in the end.
        Parameters:
            prec    precipitation variable specification, dict
                    {id, metadata, sample_times, values, flags}
        Returns:
            None
        """
        if prec['id'] not in self.prec_var_nondomain:
            self.prec_var_nondomain[prec['id']] = prec

    def _export_prec_var_nondomain(self):
        """
        Extracts the needed precipitation amount variables that could not be
        added to a multicolumn file. Export is single column here.
        This handles only non domain variables (i.e. variables which were NOT
        exported from EBAS.
        Parameters:
            None
        Returns:
            None
        """
        for _, prec in self.prec_var_nondomain.items():
            result = self.new_result()
            result.setup_non_domain()
            result.sample_times = list(prec['sample_times'])
            result._add_variable_nondomain(
                prec['id'], prec['metadata'], prec['values'], prec['flags'],
                2 if self.gen_converted else 0)
        self.prec_var_nondomain = {}



"""
nilu/argparse_helper.py
$Id: argparse_helper.py 2721 2021-10-22 23:02:49Z pe $

Module with argparse helper functions.
This module is just a helper for the default python argparse module.
It helps setting up argument parsers.

Features:
- Help Formatter that deals reasonably with newlines in help texts.
- Argument parsers for various purposes
  (datetime, interval, filename, loglevels etc)

Synopsis:


History:
V.1.0.0  2012-12.-10 pe  initial version

"""

import argparse
import logging
from . import logging_helper
import re
import textwrap
import datetime
from nilutility.datetime_helper import DatetimeInterval
from dateutil.relativedelta import relativedelta
from nilutility.datatypes import datetimeprec
from nilutility.datetime_helper import datetime_parse_iso8601

LOGLEVELNAMES = {
    'silent':   logging.CRITICAL + 1,
    'critical': logging.CRITICAL,
    'errors':   logging.ERROR,
    'warnings': logging.WARNING,
    'info+':    logging_helper.INFO_PLUS,
    'info':     logging.INFO,
    'info-':    logging_helper.INFO_MINUS,
    'debug':    logging.DEBUG}

LOGLEVELSTRING = ', '.join([k[0] for k in \
    sorted(list(LOGLEVELNAMES.items()), key=lambda lev: lev[1], reverse=True)])

class NiluArgumentParser(argparse.ArgumentParser):  # pylint: disable=R0904
    # R0904: Too many public methods
    #  --> this is due to base class argparse.ArgumentParser
    """
    Customizes the default argument parser provided by argparse.

    This class extends the ArgumentParser class by thi method:
        set_known_defaults()

    The default implementation substitutes all whitespaces (including \n)
    with a blank, which results hard to read help text for long sections.
    Implementations in RawDescriptionHelpFormatter and RawTextHelpFormatter
    go to far and do not change the format of help text at all.
    """
    def __init__(self, *args, **kwargs):
        argparse.ArgumentParser.__init__(self, *args, **kwargs)
        self.formatter_class = HelpFormatterWrapButKeepNewlines

    def set_known_defaults(self, **kwargs):
        """
        Sets defaults like the standard set_defaults method, but does __NOT__
        set defaults for unknown arguments.
        """
        for action in self._actions:
            if action.dest in kwargs:
                action.default = kwargs[action.dest]

    def add_config_arguments(self, cfgfile=True):
        """
        Add configuration relevant arguments to parser.
        Parameters:
            parser     parser object
        Returns:
            None
        """
        parser_cfg_group = self.add_argument_group('configuration options')
        if cfgfile:
            parser_cfg_group.add_argument('--cfgfile',
                                          help='configuration file to be used')

    def add_logging_arguments(self, logfile=True,
                              loglevconsole=True, def_loglevconsole='silent',
                              loglevfile=True, def_loglevfile='silent',
                              profile=True):
        # pylint: disable=R0913
        #  R0913: Too many arguments
        """
        Adds a group for logging specific arguments.
        Parameters:
            parser             existing argpase.parser object
            logfile            Bool; True if argument '--logfile' should be
                               added
            loglevconsole      Bool; True if argument '--loglevconsole' should
                               be added
            def_loglevconsole  string; default log level forconsole logging
                               ('silent', 'critical', 'errors', 'warnings',
                                'info', 'debug')
            loglevfile         Bool; True if argument '--loglevfile' should be
                               added
            def_loglevfile     string; default log level for logfiles
                               ('silent', 'critical', 'errors', 'warnings',
                                'info', 'debug')
        Returns:
            parser_log_group   argument group object
        """
        parser_log_group = self.add_argument_group('logging and debug options')
        if loglevconsole:
            parser_log_group.add_argument(
                '--loglevconsole', type=parse_loglevel_args,
                default=def_loglevconsole,
                help='log level for console output ({0})'
                .format(LOGLEVELSTRING))
        if loglevfile:
            parser_log_group.add_argument(
                '--loglevfile', type=parse_loglevel_args,
                default=def_loglevfile,
                help='log level for logfile output ({0})'
                .format(LOGLEVELSTRING))
        if logfile:
            parser_log_group.add_argument(
                '--logfile', type=str, default=None,
                help='Logfile name (including path)')
        if profile:
            parser_log_group.add_argument(
                '--profile', action='store_true',
                help='activate code profiling (developer only)')
        return parser_log_group

    def add_db_args(self, nowrite=False, nocommit=False, transcomment=False):
        """
        Add a group for DB relevant arguments to parser.
        Parameters:
            nowrite       include nowrite option
            nocommit      include nocommit option
            transcomment  include transaction comment argument
        Returns:
            None
        """
        parser_db_group = self.add_argument_group(
            'database parameters',
            textwrap.dedent('''\
            Database user and password can also be fetched from your .netrc
            file (machine=ebas_db).
            In this case do not specify --dbUser and --dbPasswd as an \
            argument.'''))
        parser_db_group.add_argument(
            '--nodb', action='store_true',
            help='do not connect to database (work offline, useful for e.g. '
                 '--version, --help etc. when the database is not reachable)')
        parser_db_group.add_argument(
            '--dbHost', type=str, default=None,
            help='db host name or IP')
        parser_db_group.add_argument(
            '--dbPort', type=int, default=None,
            help='IP port for DB connection')
        parser_db_group.add_argument(
            '--db', type=str, default=None,
            help='DB to be used.')
        parser_db_group.add_argument(
            '--dbUser', type=str, default=None,
            help='database username.')
        parser_db_group.add_argument(
            '--dbPasswd', type=str, default=None,
            help='database password. discouraged, more secure hidden input '
                 'if not set on commandline')
        if nowrite:
            parser_db_group.add_argument(
                '--nowrite', action='store_true',
                help='do not write to DB, just work on memory objects '
                '(test/debug)')
        if nocommit:
            parser_db_group.add_argument(
                '--nocommit', action='store_true',
                help='write but do not commit to DB (test/debug)')
        if transcomment:
            parser_db_group.add_argument(
                '--transcomment', type=str, default=None,
                help='transaction comment (archived in database)')


class ToggleAction(argparse.Action):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Argparse action class. Implements toggeling a boolean option by
    using the --xxx / --no-xxx / --reset-xxx logic.
    """

    def __init__(self, option_strings, dest,
                 default=None, required=False,
                 type=bool,
                 help=None):
        """
        Initialize the object.
        """
        super(ToggleAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            required=required,
            help=help)

        if len(option_strings) != 2:
            raise RuntimeError("ToggleAction needs 2 option strings")
        if sum([x.startswith('--include-') for x in option_strings]) == 1 and \
           sum([x.startswith('--exclude-') for x in option_strings]) == 1:
            # exactly 1 argument starts with --include- and one with --exclude-
            self.evaluate = self.evaluate_include_exclude
        elif sum([x.startswith('--no-') for x in option_strings]) == 1:
            # exactly 1 argument starts with --no-
            self.evaluate = self.evaluate_no_with_dash
        elif sum([x.startswith('--no') for x in option_strings]) == 1:
            # exactly 1 argument starts with --no
            self.evaluate = self.evaluate_no_without_dash
        else:
            # use the first action for true, the second for false:
            self.evaluate = lambda o: True if o == option_strings[0] else False

    def __call__(self, parser, namespace, values,  # @UnusedVariable
                 option_string=None):
        """
        Perform the action.
        """
        if option_string.startswith('--reset-'):
            setattr(namespace, self.dest, None)
        else:
            setattr(
                namespace, self.dest, self.type(self.evaluate(option_string)))

    def evaluate_include_exclude(self, option_string):
        """
        Evaluate the options string for the --include-xxx / --exclude-xxx mode
        of operation.

        Parameter:
            option_string   otpion string to be evaluated

        Returns:
            True/False
        """
        if option_string.startswith('--include-'):
            return True
        if option_string.startswith('--exclude-'):
            return False
        raise RuntimeError('option string inot recognized')

    def evaluate_no_with_dash(self, option_string):
        """
        Evaluate the options string for the --xxx / --no-xxx mode of operation.

        Parameter:
            option_string   otpion string to be evaluated

        Returns:
            True/False
        """
        if option_string.startswith('--no-'):
            return False
        else:
            return True

    def evaluate_no_without_dash(self, option_string):
        """
        Evaluate the options string for the --xxx / --noxxx mode of operation.

        Parameter:
            option_string   otpion string to be evaluated

        Returns:
            True/False
        """
        if option_string.startswith('--no'):
            return False
        else:
            return True


class ExclusiveMultiswithAction(argparse.Action):  # pylint: disable=R0903
    # R0903: Too few public methods
    """
    Argparse action class. Implements a multiple state switch.
    E.g. --red --green --blue (only one possible)
         result goes to one and the same attribute
    Important: type needs to be a class that accepts the option names as value
    """

    def __init__(self, option_strings, dest,
                 default=None, required=False,
                 type=int,
                 help=None):
        """
        Initialize the object.
        """
        super(ExclusiveMultiswithAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            required=required,
            help=help)

    def __call__(self, parser, namespace, values,  # @UnusedVariable
                 option_string=None):
        """
        Perform the action.
        """
        setattr(namespace, self.dest, self.type(option_string.lstrip('-')))


class HelpFormatterWrapButKeepNewlines(argparse.HelpFormatter):
    """
    Customizes the default help formatters provided by argparse.

    This class respects existing newline characters in the help texts, but
    additionally wraps long lines.
    The default implementation substitutes all whitespaces (including \n)
    with a blank, which results hard to read help text for long sections.
    Implementations in RawDescriptionHelpFormatter and RawTextHelpFormatter
    go to far and do not change the format of help text at all.
    """

    def _get_help_string(self, action):
        helpstr = action.help
        defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
        # pylint: disable=W0212
        # W0212: Access to a protected member _StoreTrueAction of a client
        # class
        if '(default:' in action.help or \
           (isinstance(self, argparse._StoreTrueAction) and \
            action.default == False) or\
           action.default is argparse.SUPPRESS or \
           (not action.option_strings and action.nargs not in defaulting_nargs):
            return helpstr
        helpstr += '\n(default: %(default)s)'
        return helpstr

    def _split_lines(self, text, width):
        """
        Formats the description for one argument.
        This customized formatter respects newline characters already in the
        description, but additionally wraps long lines.
        """
        _whitespace_matcher = re.compile(r'[ \t]+')
        text = _whitespace_matcher.sub(' ', text).strip()
        ret = []
        for line in text.splitlines():
            wrapped = textwrap.wrap(line, width, replace_whitespace=False)
            ret += wrapped
            if not wrapped:
                ret.append('')
        return ret

    def _fill_text(self, text, width, indent):
        """
        Formats the description for one argument.
        The default implementation substitutes all whitespaces (including \n)
        with a blank, which results hard to read help text for long sections.
        This customized formatter respects newline characters already in the
        description.
        """
        _whitespace_matcher = re.compile(r'[ \t]+')
        text = _whitespace_matcher.sub(' ', text).strip()
        return '\n'.join(
            [textwrap.fill(line, width, initial_indent=indent,
                           subsequent_indent=indent)
             for line in text.splitlines()])

def parse_loglevel_args(string):
    """
    Parser for loglevel strings.

    The commandline argument's value is parsed, and the applicable logging
    level is returned.

    Parameters:
       string   cmd-line parameter string to be checked.
    Returns:
       logging.<level> applicable level number from class logging
       (silent=CRITICAL+1).
    """
    try:
        return LOGLEVELNAMES[string.lower()]
    except KeyError:
        raise argparse.ArgumentTypeError(\
            "loglevel must be in ({})".format(LOGLEVELSTRING))


class ParseIntegers(object):
    # pylint: disable=R0903
    # R0903: Too few public methods
    #  --> this is a function factory implememnted as a class, no need for more
    #      public methods
    # Could be alternatively implemented it as a pure function factory without
    # class overhead?
    """
    Factory for creating integer object parsers

    Instances of ParseIntegers are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Keyword Arguments:
        - allow_scalar   plain integer is allowed
        - allow_list     list of integers is allowed (e.g. 1,2,3,4)
        - allow_range    integer range is allowed (e.g. 1-4)
    """

    def __init__(self, allow_scalar=True, allow_list=True, allow_range=True):
        self._scalar = allow_scalar
        self._list = allow_list
        self._range = allow_range
        # generate descriptive text of allowed types
        allowed = []
        if self._scalar:
            allowed.append("integer")
        if self._list:
            allowed.append("list of integers")
        if self._range:
            allowed.append("integer range")
        self.allowed_str = ', '.join(allowed[0:-1])
        if len(allowed) == 1:
            self.allowed_str += allowed[0]
        if len(allowed) > 1:
            self.allowed_str += ' or ' + allowed[-1]

    def __call__(self, string):
        return self.parse(string)

    def __repr__(self):
        args = '(scalar={0}, list{1}, range{2})'.format(\
                             self._scalar, self._list, self._range)
        return '%s (%s)' % (type(self).__name__, args)

    def parse(self, string):
        """
        Parser for integer, integer-list (comma separated) or
        integer-min-max (<min>-<max>) strings.

        The commandline argument's value is parsed, and the applicable values
        are returned.

        Parameters:
           string   cmd-line parameter string to be checked.
        Returns:
           int
                   in case of a single integer value
           [int,...]
                   in case of integer list
           {'min': <min>, 'max': <max>}
                   in case of min/max definition
        Raises:
            argparse.ArgumentTypeError
                    in case of argument errors
        """
        rex = re.search("^([0-9]*)-([0-9]*)$", string)
        if rex:
            if not self._range:
                raise argparse.ArgumentTypeError(\
                    'integer range is not allowed')
            mini = rex.group(1)
            maxi = rex.group(2)
            mini = int(mini) if mini != '' else None
            maxi = int(maxi) if maxi != '' else None
            return {'min': mini, 'max': maxi}
        if re.search(",", string):
            if not self._list:
                raise argparse.ArgumentTypeError(\
                    'integer list is not allowed')
            ret = []
            hstr = string
            rex = re.search("^([0-9]*),", hstr)
            while rex:
                ret.append(int(rex.group(1)))
                hstr = re.sub("^([0-9]*),", "", hstr)
                rex = re.search("^([0-9]*),", hstr)
            rex = re.search("^([0-9]*)$", hstr)
            if not rex:
                raise argparse.ArgumentTypeError(\
                    'malformed list of integers')
            ret.append(int(rex.group(1)))
            return ret
        rex = re.search("^([0-9]*)$", string)
        if rex:
            if not self._scalar:
                raise argparse.ArgumentTypeError(\
                    'integer value is not allowed')
            return int(string)
        else:
            raise argparse.ArgumentTypeError(\
                self.allowed_str + ' expected')


class ParseFloats(object):
    # pylint: disable=R0903
    # R0903: Too few public methods
    #  --> this is a function factory implememnted as a class, no need for more
    #      public methods
    # Could be alternatively implemented it as a pure function factory without
    # class overhead?
    """
    Factory for creating floating point object parsers

    Instances of ParseFloats are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Keyword Arguments:
        - allow_scalar   plain float is allowed
        - allow_list     list of floats is allowed (e.g. 1.1,2.2,3,4.99)
        - allow_range    float range is allowed (e.g. 1.1-4.4)
    """

    def __init__(self, allow_scalar=True, allow_list=True, allow_range=True):
        self._scalar = allow_scalar
        self._list = allow_list
        self._range = allow_range
        # generate descriptive text of allowed types
        allowed = []
        if self._scalar:
            allowed.append("float")
        if self._list:
            allowed.append("list of floats")
        if self._range:
            allowed.append("float range")
        self.allowed_str = ', '.join(allowed[0:-1])
        if len(allowed) == 1:
            self.allowed_str += allowed[0]
        if len(allowed) > 1:
            self.allowed_str += ' or ' + allowed[-1]

    def __call__(self, string):
        return self.parse(string)

    def __repr__(self):
        args = '(scalar={0}, list{1}, range{2})'.format(\
                             self._scalar, self._list, self._range)
        return '%s (%s)' % (type(self).__name__, args)

    def parse(self, string):
        """
        Parser for float, float-list (comma separated) or
        float-min-max (<min>-<max>) strings.

        The commandline argument's value is parsed, and the applicable values
        are returned.

        Parameters:
           string   cmd-line parameter string to be checked.
        Returns:
           float
                   in case of a single floating point value
           float,...]
                   in case of float list
           {'min': <min>, 'max': <max>}
                   in case of min/max definition
        Raises:
            argparse.ArgumentTypeError
                    in case of argument errors
        """
        rex = re.search(r"^([+-]?[0-9]*(\.[0-9]*)?)-([+-]?[0-9]*(\.[0-9]*)?)$",
                        string)
        if rex:
            if not self._range:
                raise argparse.ArgumentTypeError(\
                    'float range is not allowed')
            mini = rex.group(1)
            maxi = rex.group(2)
            mini = float(mini) if mini != '' else None
            maxi = float(maxi) if maxi != '' else None
            return {'min': mini, 'max': maxi}
        if re.search(",", string):
            if not self._list:
                raise argparse.ArgumentTypeError(\
                    'float list is not allowed')
            ret = []
            hstr = string
            rex = re.search(r"^([+-]?[0-9]*(\.[0-9]*)?),", hstr)
            while rex:
                ret.append(float(rex.group(1)))
                hstr = re.sub(r"^([+-]?[0-9]*(\.[0-9]*)?),", "", hstr)
                rex = re.search(r"^([+-]?[0-9]*(\.[0-9]*)?),", hstr)
            rex = re.search(r"^([+-]?[0-9]*(\.[0-9]*)?)$", hstr)
            if not rex:
                raise argparse.ArgumentTypeError(\
                    'malformed list of floatss')
            ret.append(float(rex.group(1)))
            return ret
        rex = re.search(r"^([+-]?[0-9]*(\.[0-9]*)?)$", string)
        if rex:
            if not self._scalar:
                raise argparse.ArgumentTypeError(\
                    'floating point value is not allowed')
            return float(string)
        else:
            raise argparse.ArgumentTypeError(\
                self.allowed_str + ' expected')


class ParseStrings(object):
    # pylint: disable=R0903
    # R0903: Too few public methods
    #  --> this is a function factory implememnted as a class, no need for more
    #      public methods
    # Could be alternatively implemented it as a pure function factory without
    # class overhead?
    """Factory for creating string object parsers

    Instances of ParseStrings are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Keyword Arguments:
        - allow_scalar   single string is allowed
        - allow_list     list of strings is allowed (e.g. 1,2,3,4)
    """

    def __init__(self, allow_scalar=True, allow_list=True, allowed_values=None):
        self._scalar = allow_scalar
        self._list = allow_list
        # generate descriptive text of allowed types
        allowed = []
        if self._scalar:
            allowed.append("string")
        if self._list:
            allowed.append("list of strings")
        self.allowed_str = ', '.join(allowed[0:-1])
        if len(allowed) == 1:
            self.allowed_str += allowed[0]
        if len(allowed) > 1:
            self.allowed_str += ' or ' + allowed[-1]
        self.allowed_values = allowed_values

    def __call__(self, string):
        return self.parse(string)

    def __repr__(self):
        args = '(scalar={}, list={}, allowed={})'.format(
            self._scalar, self._list, self.allowed_values)
        return '%s (%s)' % (type(self).__name__, args)

    def parse(self, string):
        """
        Parser for string or string-list argument.

        The commandline argument's value is parsed, and the applicable values
        are returned.

        Parameters:
           string   cmd-line parameter string to be checked.
        Returns:
           string
                   in case of a single string value
           [string,...]
                   in case of a comma separated list of strings
        Raises:
            argparse.ArgumentTypeError
                    in case of argument errors
        """
        def _check_value(string):
            """
            Check for allowed values in string.
            Parameters:
                string   string to be checked
            Returns:
                None
            Raises:
                ArgumentTypeError if not allowed
            """
            if self.allowed_values and string not in self.allowed_values:
                raise argparse.ArgumentTypeError(
                    "'{}' is not allowed (only {})"
                    .format(string, ', '.join(self.allowed_values)))
        if re.search(",", string):
            if not self._list:
                raise argparse.ArgumentTypeError(\
                    'string list is not allowed')
            ret = []
            hstr = string
            rex = re.search("^([^,]*),", hstr)
            while rex:
                _check_value(rex.group(1))
                ret.append(rex.group(1))
                hstr = re.sub("^([^,]*),", "", hstr)
                rex = re.search("^([^,]*),", hstr)
            rex = re.search("^([^,]*)$", hstr)
            if not rex:
                raise argparse.ArgumentTypeError(\
                    'malformed list of strings')
            _check_value(rex.group(1))
            ret.append(rex.group(1))
            return ret
        rex = re.search("^([^,]*)$", string)
        if rex:
            if not self._scalar:
                raise argparse.ArgumentTypeError(\
                    'string is not allowed')
            _check_value(string)
            return string
        else:
            raise argparse.ArgumentTypeError(\
                self.allowed_str + ' expected')

class ParseEmailAddr(ParseStrings):
    # pylint: disable=R0903
    # R0903: Too few public methods
    #  --> this is a function factory implememnted as a class, no need for more
    #      public methods
    # Could be alternatively implemented it as a pure function factory without
    # class overhead?
    """Factory for creating email address parsers

    Instances of ParseEmailAddr are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Keyword Arguments:
        - allow_scalar   single string is allowed
        - allow_list     list of strings is allowed (e.g. 1,2,3,4)
    """

    def parse(self, string):
        """
        Parser for email-address or email-address-list argument.

        The commandline argument's value is parsed, and the applicable values
        are returned.

        Parameters:
           string   cmd-line parameter string to be checked.
        Returns:
           string
                   in case of a single email address
           [string,...]
                   in case of a comma separated list of email addresses
        Raises:
            argparse.ArgumentTypeError
                    in case of argument errors
        """
        def _check_value(string):
            """
            Check for allowed values in string.
            Parameters:
                string   string to be checked
            Returns:
                None
            Raises:
                ArgumentTypeError if not allowed
            """
            if not re.match(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$",
                            string):
                raise argparse.ArgumentTypeError(
                    "'{}' is not a valid email address")
            if self.allowed_values and string not in self.allowed_values:
                raise argparse.ArgumentTypeError(
                    "'{}' is not allowed (only {})"
                    .format(string, ', '.join(self.allowed_values)))
        if re.search(",", string):
            if not self._list:
                raise argparse.ArgumentTypeError(\
                    'email address list is not allowed')
            ret = []
            hstr = string
            rex = re.search("^([^,]*),", hstr)
            while rex:
                _check_value(rex.group(1))
                ret.append(rex.group(1))
                hstr = re.sub("^([^,]*),", "", hstr)
                rex = re.search("^([^,]*),", hstr)
            rex = re.search("^([^,]*)$", hstr)
            if not rex:
                raise argparse.ArgumentTypeError(\
                    'malformed list of email addresses')
            _check_value(rex.group(1))
            ret.append(rex.group(1))
            return ret
        rex = re.search("^([^,]*)$", string)
        if rex:
            if not self._scalar:
                raise argparse.ArgumentTypeError(\
                    'single email address is not allowed')
            _check_value(string)
            return string
        else:
            raise argparse.ArgumentTypeError(\
                self.allowed_str + ' expected')


class ParseKeywordsToNumber(object):
    # pylint: disable=R0903
    # R0903: Too few public methods
    #  --> this is a function factory implememnted as a class, no need for more
    #      public methods
    # Could be alternatively implemented it as a pure function factory without
    # class overhead?
    """Factory for creating object parsers for keywords to numbers

    Instances of ParseKeywordsToNumber are typically passed as type= arguments
    to the ArgumentParser add_argument() method.

    Keyword Arguments:
        - allow_list     list of strings is allowed (e.g. 1,2,3,4)
    """

    def __init__(self, allow_list=True, keywords=None):
        """
        Set up the parser object.
        Parameters:
            allow_list     list of keywords is allowed (e.g. 1,2,3,4)
                           in this case, the values are combined using bit-or
            keywords       dict of keywords to numbers
        Returns:
            None
        """
        self._list = allow_list
        self.allowed_str = "Keywords" + \
                           " or list of keywords" if self._list else ""
        self.keywords = keywords if keywords is not None else {}

    def __call__(self, string):
        return self.parse(string)

    def __repr__(self):
        args = '(list={})'.format(self._list)
        return '%s (%s)' % (type(self).__name__, args)

    def parse(self, string):
        """
        Parser for keyword or keyword-list argument.

        The commandline argument's value is parsed, and the applicable values
        are returned.

        Parameters:
           string   cmd-line parameter string to be checked.
        Returns:
           number   the respective number for the keyword
                    (bitwise or in case of list)
        Raises:
            argparse.ArgumentTypeError
                    in case of argument errors
        """
        if re.search(",", string):
            if not self._list:
                raise argparse.ArgumentTypeError(\
                    'keyword list is not allowed')
        ret = 0
        for key in string.split(','):
            try:
                ret |= self.keywords[key]
            except KeyError:
                raise argparse.ArgumentTypeError(
                    "'{}' is not allowed (only {})"
                    .format(string, ', '.join(list(self.keywords.keys()))))
        return ret


def parse_time(string):
    # pylint: disable=R0911
    #  R0911: Too many return statements
    #   -> many returns still better then endless elif-chains
    """
    Parses a single time string and returns start and end point of time period.

    Parameters:
        string    time string
    Returns:
        DatetimeInterval(datetimeprec, datetimeprec)
            start and end points of time interval (each with precision given
            at the commandline)
    Raises:
        argparse.ArgumentTypeError
                in case of argument errors
    """
    def gen_res(tim1, tim2, prec):
        """
        Result generator helper.
        """
        return DatetimeInterval(datetimeprec.fromdatetime(tim1, prec),
                                datetimeprec.fromdatetime(tim2, prec))
    try:
        tim = datetime_parse_iso8601(string)
    except ValueError:
        raise argparse.ArgumentTypeError('illegal time: ' + string)

    if len(string) > 20:
        # special case: delta depends on precision given in fractions of seconds
        n_digi = len(string)-20
        delta_s = 10**-n_digi
        return gen_res(tim, tim + datetime.timedelta(seconds=delta_s), "f")
    if len(string) > 17:
        return gen_res(tim, tim + datetime.timedelta(seconds=1), "S")
    if len(string) > 14:
        return gen_res(tim, tim + datetime.timedelta(minutes=1), "M")
    if len(string) > 11:
        return gen_res(tim, tim + datetime.timedelta(hours=1), "H")
    if len(string) > 8:
        return gen_res(tim, tim + datetime.timedelta(days=1), "d")
    if len(string) > 5:
        # special cases for month: use relativedelta
        return gen_res(tim, tim + relativedelta(months=1), "m")
    # special cases for year: use relativedelta
    return gen_res(tim, tim + relativedelta(months=12), "Y")

def parse_time_interval(string):
    """
    Parser for time intervals. Allowed are ISO dates (YYYY-MO-DDTHH:MI:SS

    The commandline argument's value is parsed, and the applicable values.

    Parameters:
       string   cmd-line parameter string to be parsed.
    Returns:
       [dattimeprec, dattimeprec]
               start, endtime   datetime objects with precision
    Raises:
        argparse.ArgumentTypeError
                through parse_time()
    """
    if not string:
        return DatetimeInterval.empty()
    reg = re.match('([0-9][0-9][0-9][0-9].*)[-,]([0-9][0-9][0-9][0-9].*$)',
                   string)
    if reg:
        # FROM and TO defined
        from_time = parse_time(reg.group(1))
        to_time = parse_time(reg.group(2))
        if from_time >= to_time:
            raise argparse.ArgumentTypeError('illegal time interval: ' +\
                string + ': FROM time must be earlier than TO time')
        return DatetimeInterval(from_time[0], to_time[1])
    return parse_time(string)

def parse_time_instant(string):
    """
    Parses a time string and returns a datetime object with precision.

    Parameters:
        string    time string
    Returns:
        datetimeprec
                  timestamp with precision
    Raises:
        argparse.ArgumentTypeError
                in case of argument errors
    """
    return parse_time(string)[0]

def parse_date(string):
    """
    Parses a date string (YYYY-MO-DD) and returns a datetime.date object.

    Parameters:
        string    time string
    Returns:
        datetime.date
                  date
    Raises:
        argparse.ArgumentTypeError
                in case of argument errors
    """
    try:
        return datetime.datetime.strptime(string, '%Y-%m-%d').date()
    except ValueError:
        raise argparse.ArgumentTypeError('illegal date: ' + string)




"""
$Id: ca.py 2466 2020-07-10 13:20:48Z pe $

EBAS Masterdata Class for calibration scales

This module implements the class EbasMasterCA.
"""

from .offline_masterdata import CAOfflineMasterData
from .base import EbasMasterBase
from .co import PTR_MS_COMP

class EbasMasterCA(EbasMasterBase, CAOfflineMasterData):
    """
    Domain Class for calibration scale masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling sensor types and checking them against master data.
    Sensor type master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for SE (sensor types)
    # Those are fallback values, will be read from database as soon as possible.
    CAOfflineMasterData.read_pickle_file()

    # some hardcoded exceptional components (see method exceptional):            
    CA_EXCEPTIONAL = {
        '0': {}
    }
    # PTR-MS lev0:
    for comp in PTR_MS_COMP:
        CA_EXCEPTIONAL['0'][(comp, 'NPL')] = {
            'CO_COMP_NAME': comp,
            'CA_CALIBRATION_SCALE': 'NPL',
            'CA_DESC': None,
            'CA_COMMENT': None,
        }


    def __init__(self, dbh=None):
        """
        Read se masterdata if dbh is provided.
        """
        CAOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)

    def __getitem__(self, key):
        """
        Allows dictionary like access to metadata.
        Exception for CA:
            - exceptional metadata lookup (non DB, data level dependent)
        Parameters:
            key    tuple (comp_name, calibration_scale, data_level)
                   Only the first 2 aer used for regulat lookup!
                   data_level is optional and is used for finding exceptional
                   data level dependent masterdata
        """
        # first try: use regular masterdata, ignore data level
        try:
            return self.__class__.META[key[:2]]
        except KeyError:
            # don't exist: try exceptional masterdata (needs data level)
            # Those masterdata are NOT defined in the database, but might be
            # used e.g. in lev 0 files.
            # Thus they are accepted when reading the file, but the domain
            # layer will issue an error message.
            if len(key) != 3:
                raise
            data_level = key[2]
            if data_level in ('0a', '0b'):
                data_level = '0'
            return self.__class__.CA_EXCEPTIONAL[data_level][key[:2]]

    @classmethod
    def exist_calibrationscale(cls, cal_scale):
        """
        Check if the sensor type exists (regardless of FT and CO).
        Parameters:
            cal_scale    sensor type code to be checked
        Returns:
            True/False
        """
        if any([key[1] == cal_scale for key in list(cls.META.keys())]):
            return True
        if any([key[1] == cal_scale
                for dl_ in cls.CA_EXCEPTIONAL.keys()
                for key in list(cls.CA_EXCEPTIONAL[dl_].keys())]):
            return True
        return False
